from enum import Enum, IntEnum


class CommonPorts(Enum):
    """Common ports in every module.
    
    INWARDS : All input variables not belonging to a physical port
    OUTWARDS : All output variables not belonging to a physical port
    """

    INWARDS = "inwards"
    OUTWARDS = "outwards"


class PortType(Enum):
    """Enumeration for the port type.

    A `Port` can be either `IN` or `OUT` depending if its variables are needed for calculating the
    `System` (`IN`) or results of the calculation (`OUT`).
    """

    IN = "in"
    OUT = "out"


class Scope(IntEnum):
    """Enumeration of variable scope."""

    PRIVATE = 1
    PROTECTED = 0
    PUBLIC = -1


class Validity(IntEnum):
    """Enumeration for the validity of a variable value.
    
    A valid variable value is `OK`.
    If the value is outside the limits, status is `ERROR`.
    If the value falls between the validity range and the limits, status is `WARNING`.
    """
    ERROR = -1
    WARNING = 0
    OK = 1


class RangeType(IntEnum):
    """Enumeration for the type of valid_range and limit of variable.
    
    If the value is a (lower, upper) tuple, type is `VALUE`.
    If the value is a tuple of tuples, type is `TUPLE`.
    Tf the value is None, type is `NONE`.
    """
    NONE = 0
    VALUE = 1
    TUPLE = 2




"""Originally taken from http://code.activestate.com/recipes/576696/"""

import collections
from weakref import proxy


class Link:
    __slots__ = 'prev', 'next', 'elem', '__weakref__'


class OrderedSet(collections.abc.MutableSet):
    """A set of unique items that preserves the insertion order.

    Created by: Raymond Hettinger
    Version: 2009-03-19
    Licence: http://opensource.org/licenses/MIT
    Source: http://code.activestate.com/recipes/576694/

    An ordered set is implemented as a wrapper class for a dictionary
    implementing a doubly linked list. It also has a pointer to the last item
    in the set (self.end) which is used by the add and _iterate methods to add 
    items to the end of the list and to know when an iteration has finished,
    respectively."""
    # Big-O running times for all methods are the same as for regular sets.
    # The internal self.__map dictionary maps elems to links in a doubly linked list.
    # The circular doubly linked list starts and ends with a sentinel element.
    # The sentinel element never gets deleted (this simplifies the algorithm).
    # The prev/next links are weakref proxies (to prevent circular references).
    # Individual links are kept alive by the hard reference in self.__map.
    # Those hard references disappear when a elem is deleted from an OrderedSet.

    def __init__(self, iterable=None):
        self.__root = root = Link()  # sentinel node for doubly linked list
        root.prev = root.next = root
        self.__map = {}  # elem --> link
        if iterable is not None:
            self |= iterable

    def __len__(self) -> int:
        return len(self.__map)

    def __contains__(self, elem):
        return elem in self.__map

    def add(self, elem):
        """Append element as a new link at the end of the linked list"""
        if elem not in self.__map:
            self.__map[elem] = link = Link()
            root = self.__root
            last = root.prev
            link.prev, link.next, link.elem = last, root, elem
            last.next = root.prev = proxy(link)

    def append(self, elem):
        """Append an element to the set (mimicks the behaviour of list.append)."""
        self.add(elem)

    def index(self, value):
        """Returns the position of `elem` (mimicks the behaviour of list.index)."""
        for i, elem in enumerate(self):
            if elem == value:
                return i
        raise ValueError(f"{value!r} is not in set")

    def discard(self, elem):
        """Remove an element from the set"""   
        if elem in self.__map:
            link = self.__map.pop(elem)
            link.prev.next = link.next
            link.next.prev = link.prev

    def insert(self, index, elem):
        """Store new element at a given position in the linked list"""
        if elem in self.__map:
            return
        index = self._insertion_index(index)
        curr = self.__root
        for i in range(index):
            curr = curr.next
        self.__map[elem] = link = Link()
        link.prev, link.next, link.elem = curr, curr.next, elem
        curr.next = proxy(link)

    def __iter__(self):
        """Iterate over the linked list in order."""
        root = self.__root
        curr = root.next
        while curr is not root:
            yield curr.elem
            curr = curr.next

    def __reversed__(self):
        """Iterate over the linked list in reverse order."""
        root = self.__root
        curr = root.prev
        while curr is not root:
            yield curr.elem
            curr = curr.prev

    def pop(self, last=True):
        """Pop the last element of the set if `last` is True.
        Otherwise, pop the first element."""
        if not self:
            raise KeyError('set is empty')
        elem = next(reversed(self)) if last else next(iter(self))
        self.discard(elem)
        return elem

    def __repr__(self):
        elems = list(self) if self else ""
        return f"{self.__class__.__name__}({elems})"

    def __eq__(self, other):
        return isinstance(other, OrderedSet) and list(self) == list(other)

    def _insertion_index(self, index: int) -> int:
        """Get insertion position in the linked list"""
        if not isinstance(index, int):
            raise TypeError("insertion index must be int")
        n = len(self)
        index = min(index, n)
        if index < 0:
            index = n + max(index, -n)
        return index

    @property
    def first(self):
        try:
            return self.__root.next.elem
        except AttributeError:
            raise AttributeError("empty set")

    @property
    def last(self):
        try:
            return self.__root.prev.elem
        except AttributeError:
            raise AttributeError("empty set")



import base64
import io
import json
import pickle
from typing import Any, Dict

import numpy


class JSONEncoder(json.JSONEncoder):
    """Encode CoSApp object for JSON serialization."""

    def default(self, obj):
        if isinstance(obj, bytes):
            return "data:text/plain;base64," + base64.b64encode(obj).decode("utf-8")
        elif isinstance(obj, numpy.ndarray):
            buf = io.BytesIO()
            numpy.save(buf, obj)
            return "data:application/vnd.numpy.ndarray;base64," + base64.b64encode(buf.getvalue()).decode('utf-8')
        elif hasattr(obj, '__json__'):
            return getattr(obj, '__json__')()
        else:
            try:
                pickled_data = pickle.dumps(obj, protocol=4)
            except (pickle.PicklingError, RecursionError):
                return super(JSONEncoder, self).default(obj)
            else:
                return "data:application/vnd.python3.pickle;base64," + base64.b64encode(pickled_data).decode('utf-8')


def decode_cosapp_dict(document: Dict[str, Any]) -> Dict[str, Any]:
    """Convert JSON serialization back."""

    for key, value in document.items():
        if isinstance(value, dict):
            document[key] = decode_cosapp_dict(value)
        elif isinstance(value, str):
            if value.startswith("data:"):
                meta, data = value.split(",", maxsplit=1)
                meta = meta[5:].lower()

                if meta.endswith(";base64"):
                    data = base64.b64decode(bytes(data, encoding="utf-8"))

                if meta.startswith("text/plain;"):
                    document[key] = data
                elif meta.startswith("application/vnd.numpy.ndarray;"):
                    buf = io.BytesIO(data)
                    document[key] = numpy.load(buf)
                elif meta.startswith("application/vnd.python3.pickle;"):
                    document[key] = pickle.loads(data)
                else:
                    raise ValueError(f"Unknown data type {meta}.")

    return document



import itertools
from collections import OrderedDict
from typing import Any, Dict, List, Union

import numpy

from cosapp.drivers.abstractsetofcases import AbstractSetOfCases
from cosapp.utils.helpers import check_arg


# TODO
# [ ] Quid for vector variables
class LinearDoE(AbstractSetOfCases):
    """
    This driver builds a set of linear Doe

    Parameters
    ----------
    name : str
        Name of the driver
    owner : System, optional
        :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
    **kwargs : Any
        Keyword arguments will be used to set driver options
    """

    __slots__ = ('input_vars')

    def __init__(
        self, name: str, owner: "Optional[cosapp.systems.System]" = None, **kwargs
    ) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str, optional
            Name of the `Module`
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
        **kwargs : Dict[str, Any]
            Optional keywords arguments
        """
        super().__init__(name, owner, **kwargs)
        self.input_vars = OrderedDict()  # type: collections.OrderedDict
            # desc="Variable names varying between cases"

    def add_input_var(self,
        definition: Union[str, Dict[str, Any]],
        lower: float = None,
        upper: float = None,
        count: int = 2,
    ):
        check_arg(definition, 'definition', (str, dict))

        def add_unique_input_var(name: str, lower: float, upper: float, count: int = 2):
            self.check_owner_attr(name)
            self.input_vars[name] = {"lower": lower, "upper": upper, "count": count}

        if isinstance(definition, dict):
            for key, value in definition.items():
                if isinstance(value, dict):
                    try:
                        add_unique_input_var(key, **value)
                    except TypeError:
                        raise TypeError(
                            f"'lower' and 'upper' keys expected; got {value.keys()}."
                        )
                else:
                    raise TypeError(
                        f"dictionary expected; got {type(value).__name__}."
                    )
        else:
            add_unique_input_var(definition, lower=lower, upper=upper, count=count)

    def _build_cases(self):
        cases = list()
        for range_ in self.input_vars.values():
            cases.append(numpy.linspace(range_["lower"], range_["upper"], range_["count"]))
        self.cases = list(itertools.product(*cases))

    def _precase(self, case_idx, case):
        """Hook to be called before running each case."""
        super()._precase(case_idx, case)
        for variable, value in zip(self.input_vars, case):
            self.owner[variable] = value



import numpy as np

from cosapp.core.numerics import distributions
from cosapp.ports import Port


class XPort(Port):
    def setup(self):
        self.add_variable("x")


class X2Port(Port):
    def setup(self):
        self.add_variable("x1")
        self.add_variable("x2")


class X3Port(Port):
    def setup(self):
        self.add_variable("x1")
        self.add_variable("x2")
        self.add_variable("x3")


class V1dPort(Port):
    def setup(self):
        self.add_variable("x", np.r_[1., 2., 3.])


class V2dPort(Port):
    def setup(self):
        self.add_variable(
            "x", np.r_[11., 12., 13., 21., 22., 23., 31., 32., 33.].reshape((3, 3))
        )


class TimePort(Port):
    def setup(self):
        self.add_variable("time_", 0., unit="s")
        self.add_variable("deltaTime", 0., unit="s")


class NumPort(Port):
    def setup(self):
        self.add_variable("Pt", 101325., unit="Pa")
        self.add_variable("W", 1., unit="kg/s")


class MechPort(Port):
    def setup(self):
        self.add_variable("XN", 100., unit="rpm")
        self.add_variable("PW", 0., unit="W")


class FluidState(Port):
    def setup(self):
        self.add_variable("Tt", 273.15, unit="K")
        self.add_variable("Pt", 101325., unit="Pa")


class FlowPort(Port):
    def setup(self):
        self.add_variable("W", unit="kg/s")


class FluidPort(Port):
    def setup(self):
        self.add_variable("Tt", 273.15, unit="K")
        self.add_variable("Pt", 101325., unit="Pa")
        self.add_variable("W", 0., unit="kg/s")

    def wr(self):
        return self.W * (self.Tt / 288.15) ** 0.5 / (self.Pt / 101325.)


class XPortPerturbed(Port):
    def setup(self):
        self.add_variable('x', 1., distribution=distributions.Uniform(best=-0.1, worst=0.2), valid_range=(0., 3.))



"""
Systems package of CoSApp.
"""
from cosapp.systems.system import System, ConversionType
from cosapp.systems.systemfamily import SystemFamily
from cosapp.systems.metamodels import MetaSystem
from cosapp.systems.externalsystem import TCPSystem
from cosapp.systems.processsystem import ProcessSystem


from cosapp.systems.surrogate_models.kriging import FloatKrigingSurrogate
from cosapp.systems.surrogate_models.nearest_neighbor import (
    LinearNearestNeighbor,
    WeightedNearestNeighbor,
    RBFNearestNeighbor,
)
from cosapp.systems.surrogate_models.response_surface import ResponseSurface

__all__ = [
    "System",
    "SystemFamily",
    "MetaSystem",
    "LinearNearestNeighbor",
    "RBFNearestNeighbor",
    "WeightedNearestNeighbor",
    "FloatKrigingSurrogate",
    "ResponseSurface",
    "ConversionType",
    "ProcessSystem",
    "TCPSystem",
]



import abc

from cosapp.systems import System
from cosapp.tests.library.ports import NumPort, TimePort


# =============================================================#
# PressureLoss
#   input
#       flnum_in
#   output
#       flnum_out
#   equation
#       Pt_out = Pt_in - K * Q_in**2
#       Q_out = Q_in
#   Family
#   Sys
#   0D
#
class PressureLossFamily(abc.ABC):
    def __init__(self):
        pass


class FalseSystem:
    pass


class PressureLoss0D(System, PressureLossFamily):

    def setup(self):
        self.add_input(NumPort, "flnum_in")
        self.add_inward("K", 100.)
        self.add_outward("delta_p", 0.)
        self.add_output(NumPort, "flnum_out")

    def compute(self):
        self.flnum_out.W = self.flnum_in.W
        self.flnum_out.Pt = self.flnum_in.Pt - self.K * self.flnum_in.W ** 2
        self.delta_p = self.flnum_out.Pt - self.flnum_in.Pt


class PressureLossSys(System, PressureLossFamily):

    def setup(self):
        self.add_input(NumPort, "flnum_in")
        self.add_inward("K11", 100.)
        self.add_outward("delta_p12", 0.)
        self.add_output(NumPort, "flnum_out")

    def toPressureLoss0D(self):
        Pt_out = self.flnum_out.Pt
        Pt_in = self.flnum_in.Pt
        W_out = self.flnum_out.W
        W_in = self.flnum_in.W

        self.K = (Pt_in - Pt_out) / (W_in + W_out) ** 2 * 4


# =============================================================#
# Tank
#   description
#       the flow out from the tank is limited to W_out_tank execpt when the tank is full
#   input
#       flnum_in
#   output
#       flnum_out
#   equation
#       Pt_out = Pt_in
#       if (vol <= 0.)                      the tank is umpty
#           W_out = min(W_in, W_out_tank)
#       if (vol > 0. and vol < volMax)      the tank absorb the flow
#           W_out = W_out_tank
#       if (vol >= volMax)                  the tank is full
#           W_out = max(W_in, W_out_tank)
#   Family      To be done
#   Sys         To be done
#   0D          To be done
#
class Tank(System):

    def setup(self):
        self.add_input(NumPort, "flnum_in")
        self.add_inward("volMax", 100.)
        self.add_inward("vol", 0.)
        self.add_inward("W_out_tank", 1.)
        self.add_input(TimePort, "time_", {"time_": 0., "deltaTime": 0.})
        self.add_output(NumPort, "flnum_out")

    def compute(self):
        self.flnum_out.Pt = self.flnum_in.Pt

        W_in = self.flnum_in.W
        W_out = self.flnum_out.W

        volMax = self.volMax

        vol = self.vol
        W_out_tank = self.W_out_tank

        deltaTime = self.time_.deltaTime
        time = self.time_.time_

        if vol <= 0.:
            W_out = min(W_in, W_out_tank)

        if 0. < vol < volMax:
            W_out = W_out_tank
            deltaTimeMax = min(vol / W_out, (volMax - vol)) / W_out

        if vol >= volMax:
            W_out = max(W_in, W_out_tank)

        if deltaTime != 0:
            vol = vol - deltaTime * (W_out - W_in)

        self.flnum_out.W = W_out
        self.vol = vol


# =============================================================#
# Tank
#   description
#       the flow out from the tank is limited to W_out_tank execpt when the tank is full
#   extension to be done
#       Family, 0D, smooth in time
#   input
#       flnum_in
#   output
#       flnum_out
#   equation
#       Pt_out = Pt_in
#       if (vol <= 0.)                      the tank is umpty
#           W_out = min(W_in, W_out_tank)
#       if (vol > 0. and vol < volMax)      the tank absorb the flow
#           W_out = W_out_tank
#       if (vol >= volMax)                  the tank is full
#           W_out = max(W_in, W_out_tank)
#   Family      To be done
#   Sys         To be done
#   0D          To be done
#
class Splitter12(System):

    def setup(self):
        self.add_input(NumPort, "flnum_in")
        self.add_output(NumPort, "flnum_out1")
        self.add_output(NumPort, "flnum_out2")
        self.add_inward("x", 0.8)

        self.add_unknown("x")

    def compute(self):
        self.flnum_out1.W = self.flnum_in.W * self.x
        self.flnum_out2.W = self.flnum_in.W * (1 - self.x)
        self.flnum_out1.Pt = self.flnum_in.Pt
        self.flnum_out2.Pt = self.flnum_in.Pt


class Mixer21(System):

    def setup(self):
        self.add_input(NumPort, "flnum_in1")
        self.add_input(NumPort, "flnum_in2")
        self.add_output(NumPort, "flnum_out")
        self.add_outward("epsilon")

        self.add_equation("epsilon == 0")

    def compute(self):
        W1 = self.flnum_in1.W
        W2 = self.flnum_in2.W
        W = W1 + W2
        self.flnum_out.W = W

        Pt1 = self.flnum_in1.Pt
        Pt2 = self.flnum_in2.Pt
        Pt = (W1 * Pt1 + W2 * Pt2) / (W1 + W2)
        self.flnum_out.Pt = Pt

        self.epsilon = (Pt2 - Pt1) / (Pt2 + Pt1)



import re
from typing import Any, List, Tuple
from cosapp.utils.helpers import check_arg


def has_time(expression: Any) -> bool:
    """Checks if an expression contains 't'"""
    return re.search(r"\bt\b", str(expression)) is not None


class NameChecker:
    """Class handling admissible names, through regular expression filtering"""
    def __init__(self,
        pattern = r"^[A-Za-z][\w]*$",
        message = "Name must start with a letter, and contain only alphanumerics and '_'",
        excluded: List[str] = [],
        ):
        self.__error_message = lambda name: None  # type: Callable[[str], str]
        self.__message = ""  # type: str
        self.__pattern = None  # type: re.Pattern
        self.__excluded = tuple()
        self.pattern = pattern
        self.message = message
        self.excluded = excluded

    @classmethod
    def reserved(cls) -> List[str]:
        """List of reserved names"""
        return ["t", "time"]

    @property
    def pattern(self) -> str:
        return self.__pattern.pattern

    @pattern.setter
    def pattern(self, pattern: str) -> None:
        self.__pattern = re.compile(pattern)

    @property
    def excluded(self) -> Tuple[str]:
        return self.__excluded

    @excluded.setter
    def excluded(self, excluded) -> None:
        excluded = excluded or []
        if isinstance(excluded, str):
            excluded = [excluded]
        else:
            check_arg(excluded, 'excluded', (list, tuple, set),
                value_ok = lambda col: all(isinstance(s, str) for s in col)
            )
        self.__excluded = tuple(excluded)

    @property
    def message(self) -> str:
        """Returns a human-readable message, transcripted from regexp rule"""
        return self.__message

    @message.setter
    def message(self, message: str) -> None:
        check_arg(message, "message", str)
        self.__message = message
        if len(self.__message) > 0:
            self.__error_message = lambda name: f"{self.__message}; got {name!r}."
        else:
            self.__error_message = lambda name: f"Invalid name {name!r}"

    def is_valid(self, name: str) -> bool:
        """Method indicating whether or not a name is valid, under the current rule"""
        try:
            self(name)
        except:
            return False
        else:
            return True

    def __call__(self, name) -> str:
        """Returns `name` if valid; otherwise, raises an exception"""
        check_arg(name, "name", str)
        message = None
        reserved = self.reserved()
        if name in reserved:
            message = f"Names {reserved} are reserved"
        elif name in self.excluded:
            message = f"Names {self.excluded} are invalid"
        elif self.__pattern.match(name) is None:
            message = self.__error_message(name)
        if message is not None:
            raise ValueError(message)
        return name



"""Pretty print functions for inputs and outputs."""
from collections import OrderedDict
import logging
import sys
from typing import Any, Dict, List, Tuple

import numpy

from cosapp.systems import System

logger = logging.getLogger(__name__)

# Default value use in list_inputs method to be able to tell if the caller set a value to out_stream
_DEFAULT_OUT_STREAM = object()


def list_inputs(
    system: System,
    values: bool = True,
    inwards: bool = False,
    hierarchical: bool = True,
    print_arrays: bool = False,
    itself: bool = False,
    out_stream=_DEFAULT_OUT_STREAM,
) -> List[Tuple[str, Dict[str, Any]]]:
    """
    Return and optionally log a list of input names and other optional information.

    Parameters
    ----------
    system : System
        System to be listed
    values : bool, optional
        When True, display/return input values. Default is True.
    inwards : bool, optional
        When True, display/return inwards values. Default is False.
    hierarchical : bool, optional
        When True, human readable output shows variables in hierarchical format.
    print_arrays : bool, optional
        When False, in the columnar display, just display norm of any ndarrays with size > 1.
        The norm is surrounded by vertical bars to indicate that it is a norm.
        When True, also display full values of the ndarray below the row. Format is affected
        by the values set with numpy.set_printoptions
        Default is False.
    itself : bool, optional
        When True, include own inwards, outwards and port in the list
    out_stream : file-like object
        Where to send human readable output. Default is sys.stdout.
        Set to None to suppress.

    Returns
    -------
    list
        list of input names and other optional information about those inputs
    """
    if not values and not inwards:
        logger.warning("Expect at least values or inwards to be True.")

    inputs = get_list_inputs(system, values=values, inwards=inwards, itself=itself)
    if out_stream == _DEFAULT_OUT_STREAM:
        out_stream = sys.stdout

    if out_stream:
        _write_outputs("input", inputs, hierarchical, print_arrays, out_stream)

    return inputs


def list_outputs(
    system: System,
    values: bool = True,
    local: bool = False,
    residuals: bool = False,
    hierarchical: bool = True,
    print_arrays: bool = False,
    itself: bool = False,
    out_stream=_DEFAULT_OUT_STREAM,
) -> List[Tuple[str, Dict[str, Any]]]:
    """
    Return and optionally log a list of input names and other optional information.

    Parameters
    ----------
    system : System
        System to be listed
    values : bool, optional
        When True, display/return input values. Default is True.
    local : bool, optional
        When True, display/return local values. Default is False.
    residuals : bool, optional
        When True, display/return residual values. Default is False.
    hierarchical : bool, optional
        When True, human readable output shows variables in hierarchical format.
    print_arrays : bool, optional
        When False, in the columnar display, just display norm of any ndarrays with size > 1.
        The norm is surrounded by vertical bars to indicate that it is a norm.
        When True, also display full values of the ndarray below the row. Format is affected
        by the values set with numpy.set_printoptions
        Default is False.
    itself : bool, optional
        When True, include own inwards, outwards and port in the list
    out_stream : file-like object
        Where to send human readable output. Default is sys.stdout.
        Set to None to suppress.

    Returns
    -------
    list
        list of output names and other optional information about those outputs
    """
    if not values and not local and not residuals:
        logger.warning("Expect at least values or inwards or residuals to be True.")

    outputs = get_list_outputs(
        system, values=values, local=local, residuals=residuals, itself=itself
    )
    if out_stream == _DEFAULT_OUT_STREAM:
        out_stream = sys.stdout

    if out_stream:
        _write_outputs("output", outputs, hierarchical, print_arrays, out_stream)

    return outputs


def get_list_outputs(
    system: System,
    values: bool = True,
    residuals: bool = False,
    local: bool = False,
    parent_name: str = "",
    itself: bool = False,
) -> List[Tuple[str, Dict[str, Any]]]:
    """Return and optionally log a list of input names and other optional information.

    Parameters
    ----------
    system : System
        System to be listed
    values : bool, optional
        When True, take into account input values. Default is True.
    residuals : bool, optional
        When True, take into account residual values. Default is False.
    local : bool, optional
        When True, display/return local variables. Default is False.
    parent_name : string, optional
        Where to store parent name during recursion
    itself : bool, optional
        When True, include own inwards, outwards and port in the list
    Returns
    -------
    list_out
        list of output names and other optional information about those outputs
    """
    path_name = lambda *args: parent_name + ".".join(args)

    list_out = list()
    if len(system.children) == 0 or itself:
        for name_port, port in system.outputs.items():
            outward = (name_port == System.OUTWARDS)
            if local and outward:
                for name_val in port:
                    val = port[name_val]
                    if isinstance(val, (int, float, numpy.float64)):
                        outs = {"value": val}
                        if residuals:
                            outs["resids"] = ""
                        # if unit:  # TODO
                        #     outs['unit'] = meta[name]['unit']
                        pathname = path_name(system.name, name_val)
                        list_out.append((pathname, outs))
            if values and not outward:
                for name_val in port:
                    val = port[name_val]
                    if isinstance(val, (int, float, numpy.float64)):
                        outs = {"value": val}
                        if residuals:
                            outs["resids"] = ""
                        # if unit:  # TODO
                        #     outs['unit'] = meta[name]['unit']
                        pathname = path_name(system.name, port.name, name_val)
                        list_out.append((pathname, outs))
        if residuals:
            for residue, obj in system.residues.items():
                name_val = residue
                val = obj.value
                if isinstance(val, (int, float, numpy.float64)):
                    outs = {"resids": val}
                    if values:
                        outs["value"] = ""
                    # if unit:  # TODO
                    #     outs['unit'] = meta[name]['unit']
                    pathname = path_name(system.name, "residue", name_val)
                    list_out.append((pathname, outs))

    parent_name = parent_name + system.name + "."
    comp = [child for child in system.children]
    for child in system.exec_order:
        # if child in system.exec_order:  # respect the executive order in priority
        c = system.children[child]
        list_out.extend(
            get_list_outputs(
                c,
                values=values,
                local=local,
                residuals=residuals,
                parent_name=parent_name,
                itself=itself,
            )
        )
    for child in comp:
        if child not in system.exec_order:
            c = system.children[child]
            list_out.extend(
                get_list_outputs(
                    c,
                    values=values,
                    local=local,
                    residuals=residuals,
                    parent_name=parent_name,
                    itself=itself,
                )
            )
    return list_out


def get_list_inputs(
    system: System,
    values: bool = True,
    inwards: bool = False,
    parent_name: str = "",
    with_unfrozen_data: bool = True,
    itself: bool = False,
) -> List[Tuple[str, Dict[str, Any]]]:
    """Return and optionally log a list of input names and other optional information.

    Parameters
    ----------
    system : System
        System to be listed
    values : bool, optional
        When True, take into account input values. Default is True.
    inwards : bool, optional
        When True, take into account inwards values. Default is False
    parent_name : string, optional
        Where to store parent name during recursion
    with_unfrozen_data: bool, optional
        When True, include the unfrozen inwards. Default is True
    itself : bool, optional
        When True, include own inwards, outwards and port in the list
    Returns
    -------
    list_in
        list of input names and other optional information about those inputs
    """
    path_name = lambda *args: parent_name + ".".join(args)

    list_in = list()
    if len(system.children) == 0 or itself:
        for name_port, port in system.inputs.items():
            if inwards and name_port == System.INWARDS:
                for name_val in port:
                    val = port[name_val]
                    # if name_val not in ['data.map', 'data.map_file']:
                    if isinstance(val, (int, float, numpy.float64)):
                        outs = {"value": val}
                        # if unit:  #TODO
                        #     outs['unit'] = meta[name]['unit']
                        pathname = path_name(system.name, name_val)
                        list_in.append((pathname, outs))
                    # TODO something not right in the following two lines
                    if not with_unfrozen_data and not port.is_frozen(val):
                        list_in.pop(pathname)
            if values and name_port != System.INWARDS:
                for name_val in port:
                    val = port[name_val]
                    if isinstance(val, (int, float, numpy.float64)):
                        outs = {"value": val}
                        # if unit:  #TODO
                        #     outs['unit'] = meta[name]['unit']
                        pathname = path_name(system.name, port.name, name_val)
                        list_in.append((pathname, outs))

    parent_name = parent_name + system.name + "."
    comp = [child for child in system.children]
    for child in system.exec_order:
        c = system.children[child]
        list_in.extend(
            get_list_inputs(
                c,
                values=values,
                inwards=inwards,
                parent_name=parent_name,
                itself=itself,
            )
        )
    for child in comp:
        if child not in system.exec_order:
            c = system.children[child]
            list_in.extend(
                get_list_inputs(
                    c,
                    values=values,
                    inwards=inwards,
                    parent_name=parent_name,
                    itself=itself,
                )
            )
    return list_in


def _write_outputs(
    in_or_out: str, outputs: list, hierarchical: bool, print_arrays: bool, out_stream
) -> None:
    """
    Write table of variable names, values, residuals, and metadata to out_stream.

    The output values could actually represent input variables.
    In this context, outputs refers to the data that is being logged to an output stream.

    Parameters
    ----------
    in_or_out : str, {'input', 'output'}
        indicates whether the values passed in are from inputs or output variables.
    outputs : list
        list of (name, dict of vals and metadata) tuples.
    hierarchical : bool
        When True, human readable output shows variables in hierarchical format.
    print_arrays : bool
        When False, in the columnar display, just display norm of any ndarrays with size > 1.
        The norm is surrounded by vertical bars to indicate that it is a norm.
        When True, also display full values of the ndarray below the row. Format  is affected
        by the values set with numpy.set_printoptions
        Default is False.
    out_stream : file-like object
        Where to send human readable output.
        Set to None to suppress.
    """

    # Formatting parameters for _write_outputs and _write_output_row methods
    #     that are used by list_inputs and list_outputs methods
    _column_widths = {
        "value": 20,
        "resids": 20,
        "unit": 10,
        "shape": 10,
        "lower": 20,
        "upper": 20,
        "ref": 20,
        "ref0": 20,
        "res_ref": 20,
    }
    _align = ""
    _column_spacing = 2
    _indent_inc = 2

    if out_stream is None:
        return

    # Make a dict of outputs. Makes it easier to work with in this method
    dict_of_outputs = OrderedDict()
    for name, vals in outputs:
        dict_of_outputs[name] = vals

    count = len(dict_of_outputs)
    # Write header
    pathname = "model"
    header_name = "Input" if in_or_out == "input" else "Output"
    if in_or_out == "input":
        header = "%d %s(s) in '%s'" % (count, header_name, pathname)
    else:
        header = "%d %s(s) in '%s'" % (count, header_name, pathname)
    out_stream.write(header + "\n")
    out_stream.write("-" * len(header) + "\n" + "\n")

    if not count:
        return

    # Need an ordered list of possible output values for the two cases: inputs and outputs
    #  so that we do the column output in the correct order
    if in_or_out == "input":
        out_types = ("value", "unit")
    else:
        out_types = (
            "value",
            "resids",
            "unit",
            "shape",
            "lower",
            "upper",
            "ref",
            "ref0",
            "res_ref",
        )
    # Figure out which columns will be displayed
    # Look at any one of the outputs, they should all be the same
    outputs = dict_of_outputs[list(dict_of_outputs)[0]]
    column_names = []
    for out_type in out_types:
        if out_type in outputs:
            column_names.append(out_type)
    top_level_system_name = "top"
    # Find with width of the first column in the table
    #    Need to look through all the possible varnames to find the max width
    max_varname_len = max(len(top_level_system_name), len("varname"))
    if hierarchical:
        for name in dict_of_outputs:
            for i, name_part in enumerate(name.split(".")):
                total_len = (i + 1) * _indent_inc + len(name_part)
                max_varname_len = max(max_varname_len, total_len)
    else:
        for name in dict_of_outputs:
            max_varname_len = max(max_varname_len, len(name))

    # Determine the column widths of the data fields by finding the max width for all rows
    for column_name in column_names:
        _column_widths[column_name] = len(
            column_name
        )  # has to be able to display name!
    # for name in _var_allprocs_abs_names[in_or_out]:
    #     if name in dict_of_outputs:
    for name in dict_of_outputs:
        for column_name in column_names:
            if (
                isinstance(dict_of_outputs[name][column_name], numpy.ndarray)
                and dict_of_outputs[name][column_name].size > 1
            ):
                out = f"|{numpy.linalg.norm(dict_of_outputs[name][column_name])}|"
            else:
                out = str(dict_of_outputs[name][column_name])
            _column_widths[column_name] = max(
                _column_widths[column_name], len(str(out))
            )
    # Write out the column headers
    column_header = "{:{align}{width}}".format(
        "varname", align=_align, width=max_varname_len
    )
    column_dashes = max_varname_len * "-"
    for column_name in column_names:
        column_header += _column_spacing * " "
        column_header += "{:{align}{width}}".format(
            column_name, align=_align, width=_column_widths[column_name]
        )
        column_dashes += _column_spacing * " " + _column_widths[column_name] * "-"
    out_stream.write(column_header + "\n")
    out_stream.write(column_dashes + "\n")

    # Write out the variable names and optional values and metadata
    if hierarchical:
        out_stream.write(top_level_system_name + "\n")

        cur_sys_names = []
        # _var_allprocs_abs_names has all the vars across all procs in execution order
        #   But not all the values need to be written since, at least for output vars,
        #      the output var lists are divided into explicit and implicit
        # for varname in _var_allprocs_abs_names[in_or_out]:
        #     if varname not in dict_of_outputs:
        #         continue
        for varname in dict_of_outputs:
            # For hierarchical, need to display system levels in the rows above the
            #   actual row containing the var name and values. Want to make use
            #   of the hierarchies that have been written about this.
            existing_sys_names = []
            varname_sys_names = varname.split(".")[:-1]
            for i, sys_name in enumerate(varname_sys_names):
                if varname_sys_names[: i + 1] != cur_sys_names[: i + 1]:
                    break
                else:
                    existing_sys_names = cur_sys_names[: i + 1]

            # What parts of the hierarchy for this varname need to be written that
            #   were not already written above this
            remaining_sys_path_parts = varname_sys_names[len(existing_sys_names) :]

            # Write the Modules in the var name path
            indent = len(existing_sys_names) * _indent_inc
            for i, sys_name in enumerate(remaining_sys_path_parts):
                indent += _indent_inc
                out_stream.write(indent * " " + sys_name + "\n")
            cur_sys_names = varname_sys_names

            indent += _indent_inc
            row = "{:{align}{width}}".format(
                indent * " " + varname.split(".")[-1],
                align=_align,
                width=max_varname_len,
            )
            _write_outputs_rows(
                out_stream, row, column_names, dict_of_outputs[varname], print_arrays
            )
    else:
        for name in dict_of_outputs:
            row = "{:{align}{width}}".format(name, align=_align, width=max_varname_len)
            _write_outputs_rows(
                out_stream, row, column_names, dict_of_outputs[name], print_arrays
            )
    out_stream.write(2 * "\n")


def _write_outputs_rows(
    out_stream,
    row: str,
    column_names: List[str],
    dict_of_outputs: Dict[str, Any],
    print_arrays: bool,
):
    """
    For one variable, write name, values, residuals, and metadata to out_stream.

    Parameters
    ----------
    out_stream : file-like object
        Where to send human readable output.
        Set to None to suppress.
    row : str
        The string containing the contents of the beginning of this row output.
        Contains the name of the System or varname, possibley indented to show hierarchy.

    column_names : list of str
        Indicates which columns will be written in this row.

    dict_of_outputs : dict
        Contains the values to be written in this row. Keys are columns names.

    print_arrays : bool
        When False, in the columnar display, just display norm of any ndarrays with size > 1.
        The norm is surrounded by vertical bars to indicate that it is a norm.
        When True, also display full values of the ndarray below the row. Format  is affected
        by the values set with numpy.set_printoptions
        Default is False.

    """
    # Formatting parameters for _write_outputs and _write_output_row methods
    #     that are used by list_inputs and list_outputs methods
    _column_widths = {
        "value": 20,
        "resids": 20,
        "unit": 10,
        "shape": 10,
        "lower": 20,
        "upper": 20,
        "ref": 20,
        "ref0": 20,
        "res_ref": 20,
    }
    _align = ""
    _column_spacing = 2
    _indent_inc = 2

    if out_stream is None:
        return
    have_array_values = []  # keep track of which values are arrays
    for column_name in column_names:
        row += _column_spacing * " "
        if (
            isinstance(dict_of_outputs[column_name], numpy.ndarray)
            and dict_of_outputs[column_name].size > 1
        ):
            have_array_values.append(column_name)
            out = f"|{numpy.linalg.norm(dict_of_outputs[column_name])}|"
        else:
            out = str(dict_of_outputs[column_name])
        row += "{:{align}{width}}".format(
            out, align=_align, width=_column_widths[column_name]
        )
    out_stream.write(row + "\n")
    if print_arrays:
        left_column_width = len(row)
        spacing = left_column_width * " "
        for column_name in have_array_values:
            out_stream.write(f"{spacing}  {column_name}:\n")
            out_str = str(dict_of_outputs[column_name])
            indented_lines = [
                (left_column_width + _indent_inc) * " " + s
                for s in out_str.splitlines()
            ]
            out_stream.write("\n".join(indented_lines) + "\n")



import json
import logging
import os
from contextlib import contextmanager
from io import StringIO
from pathlib import Path
from tempfile import gettempdir
from unittest import TestCase, mock

import pytest
from jsonschema import ValidationError

from cosapp.core import config
from cosapp.core.config import CoSAppConfiguration

here = os.path.dirname(os.path.abspath(config.__file__))


@pytest.fixture()
def fakeschema():
    with open(os.path.join(here, "configuration_schema.json")) as f:
        schema = f.read()

    class FakeSchema:
        def __init__(self, *args, **kwargs):
            self.f = StringIO(schema)

        def __enter__(self):
            return self.f

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.f.close()

    return FakeSchema


def test_CoSAppConfiguration__init__(fakeschema):
    fake_param_str = json.dumps(
        {"userid": "id1234", "roles": [["scope1", "scope2"], ["scope3", "scope4"]]}
    )
    fake_io = StringIO(fake_param_str)
    with mock.patch("builtins.open") as mock_open:

        def fake_open(file, mode="r", **kwargs):
            if str(file).endswith("configuration_schema.json"):
                return fakeschema()

            if "r" in mode:
                return fake_io
            else:
                return mock.DEFAULT

        mock_open.side_effect = fake_open  # Emulate access to configuration file

        c = CoSAppConfiguration()
        assert c.userid == "id1234"
        assert {"scope1", "scope2"} in c.roles
        assert {"scope3", "scope4"} in c.roles


def test_CoSAppConfiguration_userid():
    c = CoSAppConfiguration()
    with pytest.raises(AttributeError):
        c.userid = "r12345"


def test_CoSAppConfiguration_roles():
    c = CoSAppConfiguration()
    with pytest.raises(AttributeError):
        c.roles = [["r12345"]]


def test_CoSAppConfiguration_validate_file(fakeschema):
    # minimal
    fake_io = StringIO(json.dumps({"userid": "a", "roles": []}))
    with mock.patch("builtins.open") as mock_open:

        def fake_open(file, mode="r", **kwargs):
            if str(file).endswith("configuration_schema.json"):
                return fakeschema()

            if "r" in mode:
                return fake_io
            else:
                return mock.DEFAULT

        mock_open.side_effect = fake_open  # Emulate access to configuration file

        c = CoSAppConfiguration()
        assert c.userid == "a"
        assert len(c.roles) == 0


@pytest.mark.parametrize(
    "configuration",
    [
        ({"userid": "", "roles": [["role1"]]}),  # userid to short
        ({"roles": [["role1"]]}),  # userid missing
        ({"userid": "abc"}),  # roles missing
        ({"userid": "", "roles": [["role1"]], "banana": 42}),  # unexpected keyword
        ({"userid": "", "roles": ["role1"]}),  # role is not an array
        ({"userid": "", "roles": [["role1", 1]]}),  # role contains none string element
        ({"userid": "", "roles": [["role1", "role2"], ["role1", "role2"]]}),  # duplicated role
        ({"userid": "", "roles": [["role1", "role2"], ["role3", "role3"]]}),  # duplicated tag in role
        ({"userid": "", "roles": [["role1", "role2"], ["role3", ""]]}),  # empty tag string
    ],
)
def test_CoSAppConfiguration_validate_file_error(fakeschema, configuration):
    fake_io = StringIO(json.dumps(configuration))
    with mock.patch("builtins.open") as mock_open:

        def fake_open(file, mode="r", **kwargs):
            if str(file).endswith("configuration_schema.json"):
                return fakeschema()

            if "r" in mode:
                return fake_io
            else:
                return mock.DEFAULT

        mock_open.side_effect = fake_open  # Emulate access to configuration file

        with pytest.raises(ValidationError):
            CoSAppConfiguration()


def test_CoSAppConfiguration_update_configuration(fakeschema):
    init = {"userid": "ab123", "roles": [["role1"]]}
    fake_param_str = json.dumps(init)
    fake_io = StringIO(fake_param_str)
    with mock.patch("builtins.open") as mock_open:

        def fake_open(file, mode="r", **kwargs):
            if str(file).endswith("configuration_schema.json"):
                return fakeschema()

            if "r" in mode:
                return fake_io
            else:
                return mock.DEFAULT

        mock_open.side_effect = fake_open  # Emulate access to configuration file

        c = CoSAppConfiguration()  # update_configuration is called during the init
        assert c.userid == "ab123"
        assert mock.call.__enter__().write('["role1"') in mock_open.return_value.mock_calls


def test_CoSAppConfiguration_config_var_env(fakeschema):
    folder = Path(gettempdir()).joinpath("dummy_cosapp")
    os.environ["COSAPP_CONFIG_DIR"] = str(folder)
    init = {"userid": "ab123", "roles": [["role1"]]}
    fake_param_str = json.dumps(init)
    fake_io = StringIO(fake_param_str)
    with mock.patch("builtins.open") as mock_open:

        def fake_open(file, mode="r", **kwargs):
            if str(file).endswith("configuration_schema.json"):
                return fakeschema()

            if "r" in mode:
                return fake_io
            else:
                return mock.DEFAULT

        mock_open.side_effect = fake_open  # Emulate access to configuration file

        CoSAppConfiguration()

        config_file = folder.joinpath(CoSAppConfiguration.CONFIG_FILE)
        for call in mock_open.calls:
            call in [mock.call(config_file, "r"), mock.call(config_file, "w")]


def test_CoSAppConfiguration___init__oserror(caplog):
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        with mock.patch("builtins.open") as mock_open:
            mock_open.side_effect = PermissionError

            CoSAppConfiguration()

    records = caplog.records
    warnings = [r.message for r in caplog.records if r.levelname == "WARNING"]
    assert "The configuration file cannot be opened, fall back to default." in warnings


def test_CoSAppConfiguration_update_configuration_oserror(caplog):
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        c = CoSAppConfiguration()

        with mock.patch("builtins.open") as mock_open:
            mock_open.side_effect = PermissionError
            c.update_configuration()

    records = caplog.records
    warnings = [r.message for r in caplog.records if r.levelname == "WARNING"]
    assert "Fail to save configuration locally." in warnings




"""Dummy call on VIS js views"""
import os
import tempfile

import pytest

from cosapp.tests.library.systems import FanComplex
from cosapp.tools.views.visjs import to_visjs


pytest.importorskip("jinja2")


def test_to_visjs():
    f = FanComplex('fan')
    with tempfile.NamedTemporaryFile(suffix='.html', delete=False) as temp:
        name = temp.name
    to_visjs(f, name)
    os.unlink(name)



"""
Tool to print the description of a CoSApp object.
"""
from typing import Union, Any

from cosapp.core.module import Module
from cosapp.ports.enum import PortType
from cosapp.ports.port import ExtensiblePort, Port
from cosapp.drivers.driver import Driver
from cosapp.systems.system import System


class DocDisplay:
    """Helper class to print nicely information about CoSApp classes.
    """

    # TODO unit tests + improve to have more common code with cosapp.notebook.display_doc & get_list_out/inputs

    def __init__(self, object_type: Union[type, Any]):
        """DocDisplay constructor.

        Documentation can only be built for object of type :py:class:`~cosapp.drivers.driver.Driver`,
        :py:class:`~cosapp.ports.port.Port` or :py:class:`~cosapp.systems.system.System`.

        Parameters
        ----------
        object_type : Any
            Class type of the object to display.
        """
        if not (
            isinstance(object_type, (ExtensiblePort, Module))
            or issubclass(object_type, (ExtensiblePort, Module))
        ):
            raise TypeError("Only Driver, Port and System are supported for display.")

        if isinstance(object_type, type):  # Instantiate an object
            if issubclass(object_type, Port):
                object_type = object_type("dummy", PortType.OUT)
            else:
                object_type = object_type("dummy")

        self._obj = object_type

    def __repr__(self) -> str:
        """Returns the documentation of the class given to this constructor
        object.

        Returns
        -------
        str
            Documentation string.
        """
        return self._repr_markdown_()

    def _repr_markdown_(self) -> str:
        """Returns the documentation of the class given to this constructor
        object.

        Returns
        -------
        str
            Documentation markdown formatted string.
        """
        object_type = type(self._obj)

        indent = 0
        stripped_doc = [f"Class: {object_type.__name__}", ""]
        if object_type.__doc__:
            stripped_doc.extend(["  Documentation", ""])
            for l in object_type.__doc__.split("\n"):
                if indent == 0:
                    stripped_l = l.lstrip()
                    if len(stripped_l):
                        indent = len(l) - len(stripped_l)
                else:
                    stripped_l = l[indent:]
                stripped_doc.append(stripped_l)

        if isinstance(self._obj, Module):
            if isinstance(self._obj, System):
                self._obj.run_once()
            stripped_doc.append(self._obj._repr_markdown_())

        elif isinstance(self._obj, Port):
            if len(self._obj):
                stripped_doc.extend(["", "  Variables", ""])
                stripped_doc.append(self._obj._repr_markdown_())

        return "\n".join(stripped_doc)

    @staticmethod
    def display_doc(object_type: Union[type, Any]) -> "DocDisplay":
        """Display information for `Driver`, `System` or `Port` in a Jupyter notebook.

        Parameters
        ----------
        object_type : Any
            Object of interest
        """
        return DocDisplay(object_type)


try:
    # If cosapp.notebook package is available use it
    from cosapp.notebook import display_doc
except ImportError:
    display_doc = DocDisplay.display_doc



"""
Module provides dispatch to different non-linear algorithms.
"""
import logging
from typing import Callable, Sequence, Union, Tuple, Optional, Dict, Any, Iterable

import numpy
from scipy.linalg import lu_factor, lu_solve, LinAlgWarning
from scipy import optimize

from cosapp.core.numerics.enum import NonLinearMethods
from cosapp.core.numerics.basics import SolverResults
from cosapp.utils.logging import LogLevel

logger = logging.getLogger(__name__)


class NumericSolver:
    """Class defining root routine."""

    @classmethod
    def root(cls,
             fun: Callable[[Sequence[float], Union[float, str], bool], numpy.ndarray],
             x0: Sequence[float],
             args: Tuple[Union[float, str]] = (),
             method: NonLinearMethods = NonLinearMethods.POWELL,
             jac=None,
             tol=None,
             callback=None,
             options: Optional[Dict[str, Any]] = None) -> Union[SolverResults, optimize.OptimizeResult]:
        """Try to cancel residues produced by `fun` starting with x0 as initial solution.

        Parameters
        ----------
        fun : Callable[[Sequence[float], Union[float, str]], numpy.ndarray[float]]
            Callable residues function
        x0 : Sequence[float]
            Initial solution guess
        args : Tuple[Union[float, str], ...]
            Additional arguments for `fun` - first one being the time reference
        method : NonLinearMethods
            Algorithm to use
        jac : Callable[[Sequence[float], Union[float, str]], numpy.ndarray[float]], optional
            Callable providing the jacobian matrix; None if not provided (default: None)
        tol : float, optional
        callback : Callable, optional
        options : dict, optional
            A dictionary of solver options. E.g. tol or max_iter

        Returns
        -------
        SolverResults or scipy.optimize.OptimizeResult
            The solution represented as a OptimizeResult object. Important attributes are: x the solution array,
            success a Boolean flag indicating if the algorithm exited successfully and message which describes the
            cause of the termination. See :py:class:`cosapp.core.numerics.basics.SolverResults` for a description
            of other attributes.
        """

        if not isinstance(x0, Iterable):
            x0 = numpy.asarray([x0])

        if method == NonLinearMethods.NR:
            return cls._cosapp_root(fun, x0, args=args, options=options)
        else:
            if 'verbose' in options:
                options.pop('verbose')
            results = optimize.root(fun, x0, args=args, method=method.value, jac=jac, tol=tol, callback=callback,
                                    options=options)
            results.jac_lup = (None, None)
            if 'jac' in results:
                try:
                    results.jac_lup = lu_factor(results.jac, check_finite=False)
                except (ValueError, LinAlgWarning) as err:  # Silent LU decomposition failure
                    logger.debug(f"Silent error: {err}")

            return results

    @classmethod
    def _cosapp_root(cls,
                     fresidues: Callable[[Sequence[float], Union[float, str], bool], numpy.ndarray],
                     x0: Sequence[float],
                     args: Tuple[Union[float, str]] = (),
                     options: Optional[Dict[str, Any]] = None) -> SolverResults:
        """Customized Newton-Raphson algorithm to solve `fresidues` starting with `x0`.

        Parameters
        ----------
        fresidues : Callable[[Sequence[float], Union[float, str]], numpy.ndarray[float]]
            Callable residues function
        x0 : Sequence[float]
            Initial solution guess
        args : Tuple[Union[float, str], ...]
            Additional arguments for `fun` - first one being the time reference
        options : dict, optional
            A dictionary of solver options. E.g. tol or max_iter

        Options
        -------
        eps : float
            Relative perturbation to compute derivative by finite differences.
        tol : float
            Convergence tolerance criteria
        factor : float, optional
            Relaxation factor
        max_iter : int
            Maximum number of iterations
        jac_lup : (ndarray[float], ndarray[int]), optional
            LU decomposition of Jacobian given as tuple (LU, perm) to reuse as initial direction
        jac : ndarray, optional
            Jacobian to reuse for partial update
        compute_jacobian : bool
            Force to update the Jacobian matrix even if the provided one is useful
        lower_bound : numpy.ndarray
            Min values for parameters iterated by solver.
        upper_bound : numpy.ndarray
            Max values for parameters iterated by solver.
        abs_step : numpy.ndarray
            Max absolute step for parameters iterated by solver.
        rel_step : numpy.ndarray
            Max relative step for parameters iterated by solver.
        history : bool, optional
            Store the history trace of the resolution; default False.
        jac_update_tol : float, optional
            Tolerance level for Jacobian matrix update, based on nonlinearity estimation; default 0.01

        Returns
        -------
        SolverResults or scipy.optimize.OptimizeResult
            The solution represented as a OptimizeResult object. Important attributes are: x the solution array,
            success a Boolean flag indicating if the algorithm exited successfully and message which describes the
            cause of the termination. See :py:class:`cosapp.core.numerics.basics.SolverResults` for a description
            of other attributes.
        """
        verbose = options.get('verbose', False)
        log_level = LogLevel.INFO if verbose else LogLevel.DEBUG
        it_solver = 0

        jac_rel_perturbation = options['eps']
        factor_ref = factor = options['factor']
        try:
            jac = numpy.asarray(options['jac'])
        except:
            jac = None
        jac_lu, piv = options.get('jac_lup', (None, None))
        calc_jac = options.get('compute_jacobian', True) or jac is None or jac.shape[1] != x0.shape[0]
        jac_update_tol = options.get('jac_update_tol', 0.01)
        recorder = options['recorder']
        history = options.get('history', False)
        trace = list()

        def calc_jacobian(x: numpy.ndarray,
                          residues: numpy.ndarray,
                          it_jac: int,
                          it_p_jac: int,
                          jac: numpy.ndarray=None,
                          r_indices_to_update=set()) -> Tuple[numpy.ndarray, int, int]:

            n = x.size
            if jac is None or jac.shape != (n, n):
                new_jac = numpy.zeros((n, n), dtype=float)
                x_indices_to_update = set(range(n))
            else:
                new_jac = jac.copy()
                ncol = new_jac.shape[1]
                x_indices_to_update = set()
                for i in r_indices_to_update:
                    x_indices_to_update.update([j for j in range(ncol) if new_jac[i, j] != 0])

            x_copy = x.copy()

            for j in x_indices_to_update:
                delta = jac_rel_perturbation
                if abs(x_copy[j]) >= jac_rel_perturbation:
                    delta = x_copy[j] * jac_rel_perturbation
                x_copy[j] += delta
                logger.debug(f"Perturb unknown {j}")
                perturbation = fresidues(x_copy, *args)
                new_jac[:, j] = (perturbation - residues) * (1 / delta)
                x_copy[j] = x[j]

            n_partial = len(x_indices_to_update)
            if n_partial == n:
                logger.log(log_level, f'Jacobian matrix: full update')
                it_jac += 1
            elif x_indices_to_update:
                it_p_jac += 1
                logger.log(log_level, 
                    f'Jacobian matrix: {n_partial} over {n} derivative(s) updated')

            return new_jac, it_jac, it_p_jac

        logger.debug("NR - Reference call")
        x = numpy.array(x0, dtype=float)
        r = fresidues(x, *args)
        r_norm = numpy.linalg.norm(r, numpy.inf)
        logger.log(log_level, f"Initial residue: {r_norm}")
        logger.log(LogLevel.FULL_DEBUG, f"Initial relaxation factor {factor}.")
        results = SolverResults()

        def check_numerical_features(name, default_value):
            if name not in options or options[name] is None or numpy.shape(options[name]) != x.shape:
                options[name] = numpy.full_like(x, default_value)

        check_numerical_features('lower_bound', -numpy.inf)
        check_numerical_features('upper_bound', numpy.inf)
        check_numerical_features('abs_step', numpy.inf)
        check_numerical_features('rel_step', numpy.inf)

        if history:
            trace.append({"x": x.copy(), "residues": r.copy()})
        if recorder is not None:
            recorder.record_state(str(it_solver), 'ok', )

        if not calc_jac:
            logger.debug('Reuse of previous Jacobian matrix')

        tol, max_iter = options['tol'], options['max_iter']
        p_jac, p_jac_tries = options['partial_jac'], options['partial_jac_tries']
        it_jac, it_p_jac = 0, 0  # number of full and partial evalutions of the Jacobian
        n_broyden_update = 0
        dr = numpy.zeros_like(r)
        dx = numpy.full_like(r, numpy.nan)

        try:
            res_index_to_update = {}
            consecutive_p_jac = 0

            while r_norm > tol and it_solver < max_iter:
                logger.log(log_level, f'Iteration {it_solver}')

                if calc_jac:
                    if consecutive_p_jac > p_jac_tries or not p_jac:
                        res_index_to_update = range(x.size)
                        consecutive_p_jac = 0
                    else:
                        consecutive_p_jac += 1

                    jac, it_jac, it_p_jac = calc_jacobian(x, r, it_jac, it_p_jac,
                                                          jac, res_index_to_update)

                    jac_lu, piv = cls.__lu_factor(jac)  # may raise an exception

                elif it_solver > 0:
                    # Good Broyden update - source: https://nickcdryan.com/2017/09/16/broydens-method-in-python/
                    logger.log(log_level, f'Broyden update')
                    n_broyden_update += 1
                    corr = numpy.outer((dr - jac.dot(dx)), dx) / dx.dot(dx)
                    jac += corr
                    jac_lu, piv = cls.__lu_factor(jac)

                it_solver += 1

                dx = -lu_solve((jac_lu, piv), r)
                # Compute relaxation factors from max absolute and relative steps
                with numpy.errstate(invalid='ignore', divide='ignore'):
                    abs_x = numpy.abs(x)
                    abs_dx = numpy.abs(dx)
                    dx_max_abs = options['abs_step']
                    dx_max_rel = numpy.where(abs_x > 0, abs_x * options['rel_step'], numpy.inf)
                    factor_abs = numpy.where(abs_dx > dx_max_abs, dx_max_abs / abs_dx, 1).min()
                    factor_rel = numpy.where(abs_dx > dx_max_rel, dx_max_rel / abs_dx, 1).min()

                factor = min(factor_rel, factor_abs, factor, 1)

                if factor < factor_ref:
                    logger.debug(f"\trelaxation factor lowered to {factor}")

                r_norm_prev = r_norm
                x_prev = x.copy()
                dx *= factor
                x += dx

                # Force solution within admissible bounds
                # Note:
                #   This is clearly wrong, and does NOT belong in a nonlinear solver.
                #   The correct way to deal with such constraints is to use an optimizer.
                #   This part was only kept for non-regression purposes, but should be
                #   removed in a future revision.
                x = numpy.maximum(x, options['lower_bound'])
                x = numpy.minimum(x, options['upper_bound'])

                new_r = fresidues(x, *args)
                r_norm = numpy.linalg.norm(new_r, numpy.inf)
                logger.log(log_level, f'Residue: {r_norm:.5g}')

                if numpy.allclose(x, x_prev, rtol=1e-14):
                    logger.log(log_level, f"Fixed point detected: x = {x}, dx = {x - x_prev}")
                    break

                # Estimate non-linearity by comparing extrapolated and actual residues
                abs_r = numpy.abs(r)
                extrapolated_r = r + jac.dot(dx)
                delta_vs_linear = numpy.abs(extrapolated_r - new_r)
                res_index_to_update = set(
                    numpy.argwhere(delta_vs_linear > jac_update_tol * abs_r).ravel()
                )
                res_index_to_update -= set(numpy.argwhere(abs_r < tol).ravel())

                r, dr = new_r, new_r - r

                if history:
                    trace.append({"x": x.copy(), "residues": r.copy()})
                    if calc_jac:
                        trace[-1]["jac"] = jac.copy()
                if recorder is not None:
                    recorder.record_state(str(it_solver), 'ok', )

                if res_index_to_update:
                    factor = factor_ref
                    calc_jac = True
                elif r_norm != 0:
                    growth_rate = 1 + 0.1 * (1 - factor) * r_norm_prev / r_norm
                    factor = min(growth_rate * factor, 1)
                    calc_jac = False
                    logger.log(LogLevel.FULL_DEBUG, f"New relaxation factor {factor}.")

        except (ValueError, LinAlgWarning) as err:
            is_nil = (jac == 0)
            zero_rows = numpy.argwhere(is_nil.all(axis=1)).flatten()
            zero_cols = numpy.argwhere(is_nil.all(axis=0)).flatten()
            results.success = False
            results.message = 'Singular {}x{} Jacobian matrix'.format(*jac.shape)
            results.jac_errors = {'unknowns': zero_cols, 'residues': zero_rows}
            results.jac = jac
            return results

        results.x = x
        results.jac = jac
        results.jac_lup = (jac_lu, piv)
        results.jac_calls = it_jac + it_p_jac
        results.fres_calls = it_solver
        results.success = (r_norm <= tol)
        
        convergence = f'Converged' if results.success else 'Not converged'
        results.message = (
            f'   -> {convergence} ({r_norm:.4e}) in {it_solver} iterations,'
            f' {it_jac} complete, {it_p_jac} partial Jacobian and {n_broyden_update}'
            f' Broyden evaluation(s)')
        results.trace = trace

        return results


    @staticmethod
    def __lu_factor(matrix: numpy.ndarray):
        lu, piv = lu_factor(matrix, check_finite=True)
        min_diag = numpy.abs(lu.diagonal()).min()
        if min_diag < 1e-14:
            raise LinAlgWarning(
                f"Quasi-singular Jacobian matrix; min diag element of U matrix is {min_diag}"
            )
        return lu, piv


root = NumericSolver.root


import logging
from unittest.mock import MagicMock

import pytest

from cosapp.tools.fmu.logging import FMUForwardHandler, Fmi2Status


@pytest.mark.parametrize(
    "level,status",
    [
        (logging.CRITICAL, Fmi2Status.fatal),
        (logging.ERROR, Fmi2Status.error),
        (logging.WARNING, Fmi2Status.warning),
        (logging.INFO, Fmi2Status.ok),
        (logging.DEBUG, Fmi2Status.ok),
    ],
)
def test_FMUForwardHandler_logging2FmiLevel(level, status):
    assert FMUForwardHandler.logging2FmiLevel(level) == status


@pytest.mark.parametrize(
    "handler_level",
    [logging.CRITICAL, logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG],
)
@pytest.mark.parametrize(
    "level, msg",
    [
        (logging.CRITICAL, "This is a fatal message"),
        (logging.ERROR, "This is a error message"),
        (logging.WARNING, "This is a warning message"),
        (logging.INFO, "This is a information message"),
        (logging.DEBUG, "This is a debug message"),
    ],
)
def test_FMUForwardHandler_emit(handler_level, level, msg):
    # Given
    fmu_log = MagicMock()
    fmu = MagicMock(log=fmu_log)

    FMUForwardHandler.add_handler(fmu, level=handler_level)

    # When
    logging.log(level, msg)

    # Then
    if level >= handler_level:
        message = fmu_log.call_args[0][0]
        assert message.endswith(msg)
        fmu_log.assert_called_once_with(
            message,
            FMUForwardHandler.logging2FmiLevel(level),
            debug=True if level == logging.DEBUG else False,
        )
    else:
        fmu_log.assert_not_called()



from numbers import Number
from typing import Any, Dict, List, Optional, Union, Tuple

import abc
import numpy

from cosapp.core.eval_str import EvalString
from cosapp.core.variableref import VariableReference
from cosapp.ports.enum import PortType, CommonPorts
from cosapp.ports.exceptions import ScopeError
from cosapp.utils.parsing import get_indices
from cosapp.utils.helpers import check_arg


class Boundary:
    """Numerical solver boundary.

    Parameters
    ----------
    context : cosapp.systems.System
        System in which the boundary is defined.
    name : str
        Name of the boundary
    mask : numpy.ndarray or None
        Mask of the values in the vector boundary.
    default : Number, numpy.ndarray or None
        Default value to set the boundary with.
    """
    def __init__(self,
        context: "cosapp.systems.System",
        name: str,
        mask: Optional[numpy.ndarray] = None,
        default: Union[Number, numpy.ndarray, None] = None,
        **kwargs
    ):
        super().__init__(**kwargs)  # for collaborative inheritance
        fullpath, mask = Boundary.parse(context, name, mask)

        self._mask = None  # type: Optional[numpy.ndarray]
        self._context = context  # type: cosapp.systems.System
        self._port = Boundary.portname(fullpath)  # type: str
        self._name = Boundary.fullname(fullpath)  # type: str
        self.__var = Boundary.varname(fullpath)  # type: str
        self._default_value = None  # type: Union[Number, numpy.ndarray, None]

        self.mask = mask
        self.set_default_value(default, self.mask)

    @staticmethod
    def parse(
        context: "cosapp.systems.System",
        name: str,
        mask: Optional[numpy.ndarray] = None
    ) -> Tuple[str, str, str, numpy.ndarray]:
        """
        Parse port and variable name from a name and its evaluation context.
        Also checks that the variable belongs to an input port.

        Parameters
        ----------
        context : cosapp.systems.System
            System in which the boundary is defined.
        name : str
            Name of the boundary
        mask : numpy.ndarray[bool] or None, optional
            Mask to apply on the variable; default is None (i.e. no mask)

        Returns
        -------
        tuple(full_path, mask)
            where full_path is a list of strings describing the full path of 
            the variable, and mask the array like mask to apply on the variable.
        """
        if not isinstance(name, str):
            raise TypeError(f"Variable name {type(name).__qualname__!r} is not a string.")

        var_name, auto_mask = get_indices(context, name)

        container = Boundary.container(context, var_name)

        # Force port name to be included in unknown name
        split_name = var_name.split(".")
        if len(split_name) < 2 or split_name[-2] != container.name:
            split_name.insert(-1, container.name)

        mask = mask if mask is not None else auto_mask

        return (split_name, mask)

    @staticmethod
    def portname(fullpath: List[str]) -> str:
        """Extract the port name from a variable fullpath
        
        Parameters
        ----------
        fullpath : List[str]
            The variable fullpath
        
        Returns
        -------
        str
            The port name
        """
        return ".".join(fullpath[:-1])

    @staticmethod
    def fullname(fullpath: List[str]) -> str:
        """Returns the full name from a variable fullpath
        
        Parameters
        ----------
        fullpath : List[str]
            The variable fullpath
        
        Returns
        -------
        str
            The full name
        """
        return ".".join(fullpath)

    @staticmethod
    def varname(fullpath: List[str]) -> str:
        """Extract the variable name from a variable fullpath
        
        Parameters
        ----------
        fullpath : List[str]
            The variable fullpath
        
        Returns
        -------
        str
            The variable name
        """
        return fullpath[-1]

    @staticmethod
    def container(context: "cosapp.systems.System", name: str) -> "cosapp.systems.VariableReference":
        """
        Checks that a given variable can be used as a boundary; if so, returns the variable container.

        Parameters
        ----------
        context : cosapp.systems.System
            System in which a boundary may be defined.
        name : str
            Name of the variable to check
        
        Returns
        -------
        cosapp.systems.VariableReference
            Variable container, if the variable is found valid (raises an exception otherwise)
        """
        from cosapp.ports.port import ExtensiblePort
        if name not in context:
            raise AttributeError(
                f"Variable {name!r} is not present in object {context.name!r}."
            )

        _, container, key = context.name2variable[name]
        if not isinstance(container, ExtensiblePort):
            raise TypeError(
                "Only variables can be used in mathematical algorithms; got {!r} in {!r}".format(
                    name, context.name)
            )

        if container.direction != PortType.IN:
            raise ValueError(
                "Only variables in input ports can be used as boundaries; got {!r} in '{}.{}'.".format(
                    name, context.name, container.name)
            )

        # Test if the type and scoping are compatible with the boundary status
        try:
            container.validate(key, context[name])
        except ScopeError:  # Type error should still be raised
            if context is not container.owner:
                # Only owner can set its variables
                raise ScopeError(
                    f"Trying to set variable {name!r} out of your scope through a boundary."
                )
        return container

    def __str__(self) -> str:
        return str(self.default_value)

    def __repr__(self) -> str:
        return f"{self._name} := {self!s}"

    @property
    def context(self) -> "cosapp.systems.System":
        """cosapp.systems.System : System in which the boundary is defined."""
        return self._context

    @property
    def port(self) -> str:
        """str : Name of the port containing the boundary."""
        return self._port

    @property
    def name(self) -> str:
        """str : Contextual name of the boundary."""
        return self._name

    @property
    def variable(self) -> str:
        """str : name of the variable accessed by the boundary."""
        return self.__var

    @property
    def mask(self) -> Optional[numpy.ndarray]:
        """numpy.ndarray or None : Mask of the values in the vector boundary."""
        return self._mask

    @mask.setter
    def mask(self, mask: Union[None, numpy.ndarray]) -> None:
        check_arg(mask, f"mask for variable {self.name!r}", (type(None), list, tuple, numpy.ndarray))
        if mask is not None:
            mask = numpy.asarray(mask)
            value_array = numpy.asarray(self.context[self.name])
            if value_array.ndim == 0:
                if mask.size != 1:
                    raise ValueError(
                        "Mask for variable {!r} should have size 1; got {} elements.".format(
                            self.name, mask.size
                        )
                    )
                if numpy.all(mask):
                    mask = None
            elif mask.shape != value_array.shape:
                raise ValueError(
                    "Mask shape {} for variable {!r} is wrong; expected {}.".format(
                        mask.shape, self.name, value_array.shape
                    )
                )
        self._mask = mask

    @property
    def value(self) -> Union[Number, numpy.ndarray]:
        """Number or numpy.ndarray: Current value of the boundary."""
        if self.mask is None:
            return self.context[self.name]
        elif numpy.any(self.mask):
            return self.context[self.name][self.mask]
        else:
            return numpy.empty(0)

    @value.setter
    def value(self, new: Union[Number, numpy.ndarray]) -> None:
        me = self.name

        if self.mask is None:
            if not numpy.array_equal(self.context[me], new):
                self.context[me] = new
        elif numpy.any(self.mask) and not numpy.array_equal(self.context[me][self.mask], new):
            self.context[me][self.mask] = new
            self.context.name2variable[me].mapping.owner.set_dirty(PortType.IN)

    @property
    def default_value(self) -> Union[Number, numpy.ndarray, None]:
        """Number, numpy.ndarray or None: default value for the boundary."""
        if self._default_value is None or self.mask is None:
            return self._default_value
        else:
            return self._default_value[self.mask]

    def set_default_value(self,
        value: Union[Number, numpy.ndarray, None],
        mask: Optional[numpy.ndarray] = None
    ) -> None:
        """Set the default value.

        Parameters
        ----------
        value : Number, numpy.ndarray or None
            Default value
        mask : numpy.ndarray[bool] or None, optional
            Mask to apply on the default value; default None (i.e. no mask)
        """
        if value is None:
            self._default_value = None
            return

        default_array = numpy.asarray(value)
        value_array = numpy.asarray(self.context[self.name])
        if mask is not None:
            if mask.shape != self.mask.shape:
                raise ValueError(
                    f"Provided mask shape {mask.shape} does not match current value shape {self.mask.shape}."
                )
            # Extend singleton in array or fill with nan masked default
            if mask.shape != default_array.shape:
                tmp = numpy.full_like(mask, numpy.nan, dtype=default_array.dtype)
                tmp[mask] = default_array
                default_array = tmp
            else:
                default_array = numpy.where(mask, default_array, value_array)
        elif self.mask is not None:
            if default_array.size == value_array[self.mask].size:
                tmp = numpy.full_like(value_array, numpy.nan)
                tmp[self.mask] = default_array
                default_array = tmp

            if default_array.ndim == 0:  # We have a single value masked
                self._default_value = None
                return
            mask = self.mask.copy()

        if value_array.shape != default_array.shape:
            raise ValueError(
                f"Default value {value} has not the same shape as value {self.value}."
            )

        if self.default_value is not None:
            old_mask = numpy.isfinite(self._default_value)

            default_array, mask = self._merge_masked_array(
                default_array, mask, self._default_value, old_mask)

            self.mask = mask

        elif mask is not None:
            self.mask = numpy.ma.make_mask(numpy.logical_or(mask, self.mask))

        self._default_value = default_array if default_array.ndim > 0 else default_array.tolist()

    @staticmethod
    def _merge_masked_array(
        value: Union[Number, numpy.ndarray],
        mask: Optional[numpy.ndarray],
        old_value: Union[Number, numpy.ndarray],
        old_mask: Optional[numpy.ndarray],
    ) -> Tuple[numpy.ndarray, Optional[numpy.ndarray]]:
        """Merge the old masked array with the new one.

        Parameters
        ----------
        value : Union[Number, numpy.ndarray]
            New value
        mask : Optional[numpy.ndarray]
            New mask
        old_value : Union[Number, numpy.ndarray]
            Old value
        old_mask : Optional[numpy.ndarray]
            Old mask

        Returns
        -------
        (numpy.ndarray, numpy.ndarray or None)
            The merged array with the resulting mask.
        """
        if mask is not None:
            if old_mask is not None:
                old_value = numpy.asarray(old_value)
                tmp = numpy.full_like(mask, numpy.nan, dtype=old_value.dtype)
                tmp[old_mask] = old_value[old_mask]
                tmp[mask] = value[mask]
                value = tmp
                mask = numpy.logical_or(old_mask, mask)
            else:  # An homogeneous value is set
                value[~mask] = old_value
                mask = None  # None specified indices are set to the old value
        return value, mask

    def set_to_default(self) -> None:
        """Set the current value with the default one."""
        if self._default_value is None:
            # This is to verbose and not understable with a classical use
            # logger.warning(f"No default value given for variable '{self.context.name}.{self.name}'. It will be skipped.")
            return

        nan_mask = numpy.isfinite(self._default_value)
        if self.mask is None:
            if numpy.all(nan_mask):
                self.value = self.default_value
            else:
                self.value[nan_mask] = self.default_value[nan_mask]
        else:
            sub_mask = nan_mask[self.mask]
            if numpy.all(sub_mask):
                self.value = self.default_value
            else:
                self.value[sub_mask] = self.default_value[sub_mask]


class Unknown(Boundary):
    """Numerical solver unknown.

    Parameters
    ----------
    context : cosapp.systems.System
        System in which the unknown is defined.
    name : str
        Name of the unknown
    lower_bound : float
        Minimum value authorized; default -numpy.inf
    upper_bound : float
        Maximum value authorized; default numpy.inf
    max_abs_step : float
        Max absolute step authorized in one iteration; default numpy.inf
    max_rel_step : float
        Max relative step authorized in one iteration; default numpy.inf
    mask : numpy.ndarray or None
        Mask of unknown values in the vector variable.

    Attributes
    ----------
    lower_bound : float
        Minimum value authorized; default -numpy.inf
    upper_bound : float
        Maximum value authorized; default numpy.inf
    max_abs_step : float
        Largest absolute step authorized in one iteration; default numpy.inf
    max_rel_step : float
        Largest relative step authorized in one iteration; default numpy.inf

    Notes
    -----
    The dimensionality of the variable should be taken into account in the bounding process.
    """

    def __init__(self,
        context: "cosapp.systems.System",
        name: str,
        # absolute_step: Number = 1.5e-8,  # TODO ?
        # relative_step: Number = 1.5e-8,  # TODO ?
        max_abs_step: Number = numpy.inf,
        max_rel_step: Number = numpy.inf,
        lower_bound: Number = -numpy.inf,
        upper_bound: Number = numpy.inf,
        # reference: Union[Number, numpy.ndarray] = 1.,  # TODO normalize unknown
        mask: Optional[numpy.ndarray] = None,
    ):
        super().__init__(context, name, mask)

        check_arg(max_abs_step, 'max_abs_step', Number, lambda x: x > 0)
        check_arg(max_rel_step, 'max_rel_step', Number, lambda x: x > 0)
        check_arg(lower_bound, 'lower_bound', Number)
        check_arg(upper_bound, 'upper_bound', Number)

        # TODO take into account the variable dimension in the constructor ?
        self.lower_bound = lower_bound  # type: Number
        self.upper_bound = upper_bound  # type: Number
        self.max_abs_step = max_abs_step  # type: Number
        self.max_rel_step = max_rel_step  # type: Number

    def __str__(self) -> str:
        try:
            return str(self.value)
        except KeyError:  # boundary does not exist in the current context
            return str(self.default_value)

    def copy(self) -> "Unknown":
        """Copy the unknown object.

        Returns
        -------
        Unknown
            Duplicated unknown
        """
        new = Unknown(
            self.context,
            self.name,
            max_abs_step=self.max_abs_step,
            max_rel_step=self.max_rel_step,
            lower_bound=self.lower_bound,
            upper_bound=self.upper_bound,
        )
        new._mask = None if self.mask is None else self.mask.copy()
        return new

    def to_dict(self) -> Dict[str, Any]:
        """Returns a JSONable representation of the unknown.
        
        Returns
        -------
        Dict[str, Any]
            JSONable representation
        """
        return {
            "context": self.context.contextual_name,
            "name": self.name,
            "max_abs_step": self.max_abs_step,
            "max_rel_step": self.max_rel_step,
            "lower_bound": self.lower_bound,
            "upper_bound": self.upper_bound,
            "mask": None if self.mask is None else self.mask.tolist()
        }


class AbstractTimeUnknown(abc.ABC):
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)  # for collaborative inheritance

    @abc.abstractproperty
    def der(self) -> EvalString:
        """Expression of the time derivative, given as an EvalString"""
        pass

    @abc.abstractproperty
    def max_time_step_expr(self) -> EvalString:
        """Expression of the maximum admissible time step, given as an EvalString."""
        pass

    @abc.abstractproperty
    def max_abs_step_expr(self) -> EvalString:
        """Expression of the maximum absolute step in one iteration, given as an EvalString."""
        pass

    @abc.abstractmethod
    def reset(self) -> None:
        """Reset transient unknown to a reference value"""
        pass

    @property
    def d_dt(self) -> Any:
        """Value of time derivative"""
        return self.der.eval()

    @property
    def max_abs_step(self) -> float:
        """float: Maximum absolute step in one iteration"""
        return self.max_abs_step_expr.eval()

    @property
    def max_time_step(self) -> float:
        """float: Maximum admissible time step in one iteration"""
        dt_max = self.max_time_step_expr.eval()
        dx_max = self.max_abs_step
        if numpy.isfinite(dx_max):
            step_based_dt = self.extrapolated_time_step(dx_max)
            dt_max = min(dt_max, step_based_dt)
        return dt_max

    @property
    def constrained(self) -> bool:
        """bool: is unknown constrained by a limiting time step?"""
        constrained = lambda expr: numpy.isfinite(expr.eval()) if expr.constant else True
        return constrained(self.max_time_step_expr) or constrained(self.max_abs_step_expr)

    def extrapolated_time_step(self, step: float) -> float:
        """
        Time step necessary to attain a variation of `step`
        at a rate given by current value of the time derivative.
        """
        rate = numpy.abs(self.d_dt)
        step = numpy.where(rate > 0, abs(step), numpy.inf)
        rate = numpy.where(rate > 0, rate, 1)
        return numpy.min(step / rate)


class TimeUnknown(Boundary, AbstractTimeUnknown):
    """Time-dependent solver unknown.

    Parameters
    ----------
    context : cosapp.systems.System
        System in which the unknown is defined.
    name : str
        Name of the unknown

    Attributes
    ----------
    max_time_step : float
        Max time step authorized in one iteration; default numpy.inf
    """

    def __init__(self,
        context: "cosapp.systems.System",
        name: str,
        der: Any,
        max_time_step: Union[Number, str] = numpy.inf,
        max_abs_step: Union[Number, str] = numpy.inf,
        pulled_from: Optional[VariableReference] = None,
    ):
        super().__init__(context, name)
        self._pulled_from = pulled_from
        self.__type = None
        self.__shape = None
        self.__dt_max = numpy.inf
        self.__dx_max = numpy.inf
        self.d_dt = der
        self.max_time_step = max_time_step
        self.max_abs_step = max_abs_step

    def __str__(self) -> str:
        try:
            return str(self.value)
        except KeyError:  # does not exist in current context
            return str(self.default_value)

    @property
    def der(self) -> EvalString:
        """Expression of time derivative, given as an EvalString"""
        return self.__der

    @AbstractTimeUnknown.d_dt.setter
    def d_dt(self, expression: Any):
        eval_string, value, dtype = self.der_type(expression, self.context)
        if self.__type is None:
            self.__type = dtype
            if dtype is numpy.ndarray:
                self.__shape = value.shape
            elif dtype is not Number:
                raise TypeError(
                    f"Derivative expressions may only be numbers or array-like collections; got '{value}'")
        elif self.__type is not dtype:
            raise TypeError(
                f"Expression '{expression!s}' is incompatible with declared type {self.__type.__qualname__}")
        if self.__shape and numpy.shape(value) != self.__shape:
            raise ValueError(
                f"Expression '{expression!s}' should be an array of shape {self.__shape}")
        self.__der = eval_string

    @property
    def max_time_step_expr(self) -> EvalString:
        """Maximum admissible time step, given as an EvalString."""
        return self.__dt_max

    @property
    def max_abs_step_expr(self) -> EvalString:
        """Maximum admissible step, given as an EvalString."""
        return self.__dx_max

    @AbstractTimeUnknown.max_time_step.setter
    def max_time_step(self, expression: Any):
        self.__dt_max = self.__positive_expr(expression, "max_time_step")

    @AbstractTimeUnknown.max_abs_step.setter
    def max_abs_step(self, expression: Any):
        self.__dx_max = self.__positive_expr(expression, "max_abs_step")

    def __positive_expr(self, expression: Any, name: str) -> EvalString:
        eval_string, value, dtype = self.der_type(expression, self.context)
        check_arg(value, name, Number)  # checks that expression is scalar
        if value <= 0 and eval_string.constant:
            # Note:
            #   If expression is context-dependent (non-constant), it may turn out to be positive at time driver execution.
            #   Therefore, an exception should only be raised for constant, non-positive expressions.
            raise ValueError(f"{name} must be strictly positive")
        return eval_string

    def copy(self) -> "TimeUnknown":
        """Copy time-dependent unknown object.

        Returns
        -------
        TimeUnknown
            Duplicated unknown
        """
        return TimeUnknown(self.context, self.name, self.der, self.max_time_step_expr)

    @staticmethod
    def der_type(expression: Any, context: "cosapp.systems.System") -> Tuple:
        """Static method to evaluate the type and default value of an expression used as time derivative"""
        if isinstance(expression, EvalString):
            eval_string = expression
        else:
            eval_string = EvalString(expression, context)
        value = eval_string.eval()
        if isinstance(value, (list, tuple, numpy.ndarray)):
            value = numpy.array(value)
            dtype = numpy.ndarray
        elif TimeUnknown.is_number(value):
            dtype = Number
        else:
            dtype = type(value)
        return eval_string, value, dtype

    @staticmethod
    def is_number(value) -> bool:
        """Is value suitable for a derivative?"""
        return isinstance(value, Number) and not isinstance(value, bool)

    @property
    def pulled_from(self) -> Optional[VariableReference]:
        """VariableReference or None: Original time unknown before pulling; None otherwise."""
        return self._pulled_from

    @Boundary.value.setter
    def value(self, new: Union[Number, numpy.ndarray]) -> None:
        super(self.__class__, self.__class__).value.fset(self, new)
        self.context.name2variable[self.name].mapping.owner.set_dirty(PortType.IN)

    def to_dict(self) -> Dict[str, Any]:
        """Returns a JSONable representation of the transient unknown.
        
        Returns
        -------
        Dict[str, Any]
            JSONable representation
        """
        return {
            "context": self.context.contextual_name,
            "name": self.name,
            "der": str(self.__der),
            "max_time_step": str(self.max_time_step_expr),
        }

    def reset(self) -> None:
        """Reset transient unknown to a reference value.
        Inactive for class TimeUnknown."""
        pass


class TimeDerivative(Boundary):
    """Explicit time derivative.

    Parameters
    ----------
    context : cosapp.systems.System
        System in which the unknown is defined.
    name : str
        Name of the variable
    source : str
        Variable such that name = d(source)/dt
    initial_value : Any
        Time derivative initial value
    """

    def __init__(self,
        context: "cosapp.systems.System",
        name: str,
        source: Any,
        initial_value: Any = None,
    ):
        super().__init__(context, name)
        self.__shape = None
        self.__previous = None
        eval_string, value, self.__type = self.source_type(source, self.context)
        if self.__type is numpy.ndarray:
            self.__shape = value.shape
        # Set source & initial value
        self.source = source
        self.initial_value = initial_value
        self.reset()

    def __str__(self) -> str:
        try:
            return str(self.value)
        except KeyError:  # does not exist in current context
            return str(self.default_value)

    def reset(self, value: Any = None) -> None:
        self.__previous = self.source
        if value is not None:
            self.initial_value = value  # NB: `value` may be an expression
        value = self.initial_value
        if value is not None:
            self.__set_value(value)

    @property
    def source_expr(self) -> EvalString:
        """Variable whose rate is evaluated, returned as an EvalString"""
        return self.__src

    @property
    def source(self) -> Union[Number, numpy.ndarray]:
        """Value of the variable whose rate is evaluated"""
        return self.__src.eval()

    @source.setter
    def source(self, expression: Any):
        self.__src = self.__parse(expression)
        if self.__previous is None:
            self.__previous = self.source

    @property
    def initial_value_expr(self) -> EvalString:
        """Initial value of time derivative, returned as an EvalString"""
        return self.__initial

    @property
    def initial_value(self) -> Union[Number, numpy.ndarray]:
        """Initial value of time derivative"""
        return self.__initial.eval()

    @initial_value.setter
    def initial_value(self, expression):
        if expression is None:
            self.__initial = EvalString(None, self.context)
        else:
            self.__initial = self.__parse(expression)

    def update(self, dt: Number) -> Number:
        """Evaluate rate-of-change of source over time interval `dt`"""
        current = self.source
        rate = (current - self.__previous) / dt  # backward finite-difference time derivative
        self.__set_value(rate)
        self.__previous = current
        return rate

    def __set_value(self, value: Union[Number, numpy.ndarray]):
        """Private setter for `value`"""
        super(self.__class__, self.__class__).value.fset(self, value)
        self.context.name2variable[self.name].mapping.owner.set_dirty(PortType.IN)

    @Boundary.value.setter
    def value(self, new: Union[Number, numpy.ndarray]) -> None:
        raise RuntimeError("Time derivatives are computed, and cannot be explicitly set")

    @staticmethod
    def source_type(expression: Any, context: "cosapp.systems.System") -> Tuple:
        """Static method to evaluate the type and default value of an expression used as rate source"""
        eval_string, value, dtype = TimeUnknown.der_type(expression, context)
        if dtype is numpy.ndarray:
            value.fill(0.0)
        else:
            value = 0.0
        return eval_string, value, dtype

    def __parse(self, expression: Any) -> EvalString:
        eval_string, value, dtype = TimeUnknown.der_type(expression, self.context)
        if self.__type is not dtype:
            raise TypeError(
                f"Expression '{expression!s}' is incompatible with declared type {self.__type.__qualname__}")
        if self.__shape and value.shape != self.__shape:
            raise ValueError(f"Expression '{expression!s}' should be an array of shape {self.__shape}")
        return eval_string

    def to_dict(self) -> Dict[str, Any]:
        """Returns a JSONable representation of the time derivative.
        
        Returns
        -------
        Dict[str, Any]
            JSONable representation
        """
        return {
            "context": self.context.contextual_name,
            "name": self.name,
            "source": str(self.source_expr),
            "initial_value": str(self.initial_value_expr),
        }



from cosapp.tools.fmu.exporter import to_fmu
from cosapp.tools.help import display_doc
from cosapp.tools.problem_viewer.problem_viewer import view_model
from cosapp.tools.views.d3js import to_d3
from cosapp.tools.views.markdown import port_to_md, system_to_md
from cosapp.tools.views.prettyprint import list_inputs, list_outputs
from cosapp.tools.views.visjs import to_visjs

__all__ = [
    "display_doc",
    "view_model",
    "list_inputs",
    "list_outputs",
    "port_to_md",
    "system_to_md",
    "to_d3",
    "to_fmu",
    "to_visjs",
]



# -*- coding: utf-8 -*-
#
# sample documentation build configuration file, created by
# sphinx-quickstart on Mon Apr 16 21:22:43 2012.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.
import os
import time
import sys
from pathlib import Path

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, str(Path.cwd().parent.resolve()))
sys.path.insert(0, str((Path.cwd() / "tutorials").resolve()))  # Required to find the tutorial Python scripts

from cosapp.core._version import __version__

# -- General configuration -----------------------------------------------------

# Load local extensions (e.g. mermaid)
sys.path.insert(0, str(Path.cwd().resolve() / "tools"))

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "nbsphinx",
    "mermaid",
    "mermaid_inheritance",
    "sphinx_copybutton",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
# source_suffix = ['.rst', '.ipynb']

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "CoSApp"
copyright = f"2018-{time.localtime().tm_year}, CoSApp Team - Safran SA"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = __version__
# The full version, including alpha/beta/rc tags.
release = __version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "**.ipynb_checkpoints", "trainings"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "cosapp_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = ["."]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

html_css_files =  ['css/font.css', 'css/kube.min.css',
  'css/kube.legenda.css', 'css/highlight.css', 'css/master.css',
  'css/kube.demo.css', 'css/custom.css']

html_js_files = [
    'js/tocbot.min.js',
]

# Add images for notebook gallery thumbnails
html_extra_path = ['nb_thumbnails', ]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = ""  # default: '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {"**": ["globaltoc.html", "relations.html", "searchbox.html"]}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "cosappdoc"

# -- Options for nbsphinx --------------------------------------------------

# Execute notebooks before conversion: 'always', 'never', 'auto' (default)
nbsphinx_execute = "always"

# Use this kernel instead of the one stored in the notebook metadata:
# nbsphinx_kernel_name = 'python3'

# List of arguments to be passed to the kernel that executes the notebooks:
# nbsphinx_execute_arguments = ['--InlineBackend.figure_formats={"png", "pdf"}']

# If True, the build process is continued even if an exception occurs:
nbsphinx_allow_errors = False

# Controls when a cell will time out (defaults to 30; use -1 for no timeout):
# nbsphinx_timeout = 60

# Default Pygments lexer for syntax highlighting in code cells:
# nbsphinx_codecell_lexer = 'ipython3'

# Input cell prompt
# nbsphinx_input_prompt = 'In [%s]'

# Output cell prompt
# nbsphinx_output_prompt = 'Out [%s]'

# If window is narrower than this, input/output prompts are on separate lines:
# nbsphinx_responsive_width = '700px'

# This is processed by Jinja2 and inserted before each notebook
# nbsphinx_prolog = r""""""

# This is processed by Jinja2 and inserted after each notebook
# nbsphinx_epilog = r""""""

# Notebook gallery thumbnails
nbsphinx_thumbnails = {
    "tutorials/01-Systems": "_images/file-code.svg",
    "tutorials/02-Ports": "_images/plug.svg",
    "tutorials/03-Drivers": "_images/play.svg",
    "tutorials/03b-Advanced-Drivers": "_images/workflow-all.svg",
    "tutorials/04-Triggers": "_images/zap.svg",
    "tutorials/05-Validation": "_images/verified.svg",
    "tutorials/06-Visibility": "_images/eye.svg",
    "tutorials/SystemSurrogates": "_images/fold.svg",
    "tutorials/10-Recorders": "_images/desktop-download.svg",
    "tutorials/FMI": "_images/package.svg",
    "tutorials/Logging": "_images/pulse.svg",
}

import nbsphinx

if nbsphinx.__version__ == '0.4.2':
    html_js_files.append(
        (
            "https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.4/require.min.js",
            {"integrity":"sha256-Ae2Vz/4ePdIu6ZyI/5ZGsYnb+m0JlOmKPjt6XZ9JJkA=", "crossorigin":"anonymous"}
        )
    )

# -- Options for other extensions --------------------------------------------------

## Include Python objects as they appear in source files
## Default: alphabetically ('alphabetical')
autodoc_member_order = "bysource"

## Default flags used by autodoc directives
autodoc_default_options = {
    "members": True,
    "member-order": "alphabetical",
    "undoc-members": True,
    "show-inheritance": True
}

## Generate autodoc stubs with summaries from code
autosummary_generate = True

napoleon_google_docstring = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ("index", "cosapp.tex", "CoSApp Documentation", "CoSApp Team", "manual")
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        "index",
        "cosapp",
        "CoSApp Documentation",
        ["CoSApp Team", ],
        1,
    )
]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "cosapp",
        "CoSApp Documentation",
        "CoSApp Team",
        "cosapp",
        "One line description of project.",
        "Miscellaneous",
    )
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'




"""
Configuration of CoSApp.
"""
import os
import sys
import json

try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError
import logging
from pathlib import Path
from typing import FrozenSet

import jsonschema

logger = logging.getLogger(__name__)


class CoSAppConfiguration:
    r"""Encapsulate CoSApp configuration parameters and its handlers.
    
    CoSApp configuration folder are by default store in 
    On Linux::

       $HOME/.cosapp.d

    On Windows::

       %USERPROFILE%\.cosapp.d

    This default configuration folder is overwritten by the environment
    variable ``COSAPP_CONFIG_DIR``.
    """

    COSAPP_CONFIG_DIR = ".cosapp.d"
    CONFIG_FILE = "cosapp_config.json"  # type: str

    def __init__(self) -> None:
        """Constructor"""
        self._userid = ""  # type: str
        self._roles = frozenset()  # type: FrozenSet[FrozenSet[str]]

        self.__load_configuration()

    def __get_config_path(self) -> str:
        folder = Path(
            os.environ.get("COSAPP_CONFIG_DIR", None) or 
            Path.home().joinpath(CoSAppConfiguration.COSAPP_CONFIG_DIR)
        )
        return folder.joinpath(CoSAppConfiguration.CONFIG_FILE)

    def __load_configuration(self) -> None:
        """Read configuration from file or generate the default.

        Raises
        ------
        OSError
            If the current platform is not recognized
        """
        fullpath = self.__get_config_path()
        warning_msg = "The configuration file cannot be opened, fall back to default."
        if os.path.isfile(fullpath):  # Load local parameters - offline connection
            try:
                parameters = self.validate_file(fullpath)
                self._userid = parameters["userid"]
                self._roles = frozenset(
                    [frozenset(role) for role in parameters["roles"]]
                )
            except (OSError, JSONDecodeError):
                logger.warning(warning_msg)
        else:
            logger.warning(warning_msg)
        self.update_configuration()  # Try to update user permission from official root

    @property
    def userid(self) -> str:
        """str : User ID"""
        return self._userid

    @property
    def roles(self) -> FrozenSet[FrozenSet[str]]:
        """FrozenSet[FrozenSet[str]] : Roles that the user can impersonate."""
        return self._roles

    @staticmethod
    def validate_file(path: str) -> dict:
        """Validate the provided file against JSON schema for configuration file.

        Parameters
        ----------
        path : str
            Absolute path to the file to be tested.

        Returns
        -------
        dict
            The dictionary read in the validated file

        Raises
        ------
        jsonschema.exceptions.ValidationError
            If the provided file does not conform to the JSON schema.
        """
        with open(
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "configuration_schema.json"
            )
        ) as fp:
            config_schema = json.load(fp)
        with open(path) as untested_file:
            params = json.load(untested_file)

        jsonschema.validate(params, config_schema)

        return params

    def update_configuration(self) -> None:
        """Update the configuration file for the current user.

        The update process ask the server about updated role.
        """
        if len(self._userid) == 0:
            self._userid = os.environ.get("USERNAME", "") or os.environ.get("USER", "")

            if not self._userid:
                raise OSError("Unable to find the user id.")

        # TODO - request reference roles source here

        # Save back the changes following the update
        parameters = {"userid": self.userid, "roles": list()}

        for role in self.roles:
            parameters["roles"].append(list(role))

        try:
            config_file = self.__get_config_path()
            config_dir = os.path.dirname(config_file)
            if not os.path.isdir(config_dir):
                os.makedirs(config_dir)

            with open(config_file, "w") as file:
                json.dump(parameters, file, sort_keys=True)
        except OSError:
            logger.warning("Fail to save configuration locally.")



from enum import Enum
import logging
from typing import List, Optional

from cosapp.systems.system import System, ConversionType
from cosapp.utils.helpers import check_arg

logger = logging.getLogger(__name__)


class SystemFamily(System):
    # TODO clarify why its a System and why System has `convert_to`
    # Is it abstract?
    """Abstract class defining a family of `System` convertible in one another."""

    __slots__ = ('family_name', 'modelings')

    def __init__(self, name: str, **kwargs):
        self.modelings = SystemFamilyModelings()  # type: SystemFamilyModelings
        self.family_name = "not_defined"  # type: str
        super(SystemFamily, self).__init__(name, **kwargs)

    def possible_conversions(self) -> List[str]:
        """Get the list of possible conversions between `System` type inside the `SystemFamily`.

        Returns
        -------
        List[str]
            List of conversion methods available
        """
        rtn = [meth for meth in dir(self) if type(self).__name__ + "_to_" in meth]
        return rtn

    def convert_to(self, to_type: str, auto: ConversionType = ConversionType.manual):
        """Convert the current `System` in the provided to_type class.

        Parameters
        ----------
        to_type : str
            Type name of the target new class
        auto : ConversionType, optional
            Type of conversion; default manual
        """
        check_arg(to_type, "to_type", str, lambda s: len(s) > 0)
        check_arg(auto, "auto", ConversionType)

        self_type = self.__class__.__name__

        if auto != ConversionType.manual:
            to_type = getattr(self.modelings, auto.value)()

        if self_type == to_type:
            raise ValueError(
                'Failed to convert "{}" #{} to #{}: same classes'.format(
                    self.name, self_type, to_type
                )
            )

        method_name = f"{self_type}_to_{to_type}"

        if method_name in self.possible_conversions():
            converter = getattr(self, method_name)
        else:
            raise ValueError(
                'Failed to convert "{}" #{} to #{}: method does not exist'.format(
                    self.name, self_type, to_type
                )
            )

        parent = self.parent
        if parent is None:
            raise ValueError(
                'Failed to convert "{}" #{} to #{}: parent system is immutable'.format(
                    self.name, self_type, to_type
                )
            )

        # Restore connections
        to_restore = list(filter(lambda c: c.sink.owner is self, parent.connectors.values()))
        new_obj = converter()
        new_obj.name = name = self.name
        # TODO this is really ugly and should be done by the parent...
        execution_idx = parent.exec_order.index(name)
        parent.pop_child(name)
        parent.add_child(new_obj, execution_index=execution_idx)
        for c in to_restore:
            parent.connect(c.sink, c.source, c.variable_mapping)

    def update_connections(self, new_system: System) -> None:
        # TODO this should be called automatically and not in the _to_ user method?
        """Update connections after conversion of this `System` into the `new_system`.

        New connections are created for input port. But for output port, the port
        reference in the existing connections are updated.

        Parameters
        ----------
        new_system : System
            The new type `System` in which the current `System` has been converted into
        """

        for name, port in self.inputs.items():
            if name in new_system.inputs:
                new_system.inputs[name].morph(port)
                port.owner = new_system
                new_system.inputs[name] = port
            else:
                logger.warning(
                    '"{}" conversion: "{}" not found in #{}'
                    ''.format(self.name, name, type(new_system).__name__)
                )

        for name, port in self.outputs.items():
            if name in new_system.outputs:
                # Convert existing output ports to keep existing connections
                new_system.outputs[name].morph(port)
                port.owner = new_system
                new_system.outputs[name] = port
            else:
                logger.warning(
                    '"{}" conversion: "{}" not found in #{}'
                    ''.format(self.name, name, type(new_system).__name__)
                )


class SystemFamilyModelings(list):
    # TODO documentation
    def __init__(self):
        super().__init__()

    def add(self, name, fidelity, cost):
        self.append((name, fidelity, cost))

    def delete(self, name):
        rtn = [idx for idx, tup in enumerate(self) if tup[0] == name]
        for idx in rtn:
            del self[idx]

    def exists(self, name):
        l = [idx for idx, tup in enumerate(self) if tup[0] == name]
        return len(l) > 0

    def best_fidelity_to_cost_ratio(self):
        self.sort(key=lambda tup: tup[2] / tup[1])
        return self[0][0]

    def highest_fidelity(self):
        self.sort(key=lambda tup: tup[1])
        return self[-1][0]

    def lowest_fidelity(self):
        self.sort(key=lambda tup: tup[1])
        return self[0][0]

    def highest_cost(self):
        self.sort(key=lambda tup: tup[2])
        return self[-1][0]

    def lowest_cost(self):
        self.sort(key=lambda tup: tup[2])
        return self[0][0]



import numpy, scipy.interpolate
import enum
import logging
from typing import Any, Dict, List

from cosapp.drivers.driver import Driver
from cosapp.core.numerics.boundary import Boundary
from cosapp.core.eval_str import AssignString
from cosapp.utils.helpers import check_arg

logger = logging.getLogger(__name__)


class Interpolator:
    """Class describing a function defined from tabulated data,
    based on interpolating function `interpolator`. Function `interpolator`
    must return a callable function x -> f(x), interpolating [y_0, ..., y_n] at
    points [x_0, ..., x_n]."""

    class Kind(enum.Enum):
        """Interpolator kind"""
        Linear = scipy.interpolate.interp1d
        CubicSpline = scipy.interpolate.CubicSpline
        Pchip = scipy.interpolate.PchipInterpolator
    
    def __init__(self, data, kind=Kind.Linear):
        self.__data = None
        self.kind = kind
        self.data = data

    @property
    def kind(self) -> Kind:
        return self.__kind

    @kind.setter
    def kind(self, kind: Kind) -> None:
        check_arg(kind, 'kind', self.Kind)
        self.__kind = kind
        self.__update_evaluator()

    @property
    def data(self) -> numpy.ndarray:
        return self.__data
    
    @data.setter
    def data(self, data) -> None:
        data = numpy.asarray(data, dtype=float)
        shape = data.shape
        if data.ndim != 2:
            raise ValueError(f"invalid shape {shape}; `data` must be a 2D array.")
        if shape[1] == 2:
            # data given as [[x0, y0], ..., [xn, yn]]
            data = data.T
        # sort data to ensure increasing x
        self.__data = data[:, data[0].argsort()]
        self.__data.setflags(write=0)
        self.__update_evaluator()

    def __update_evaluator(self) -> None:
        data = self.data
        if data is None:
            def evaluator(x):
                raise ValueError("data was not specified")
        else:
            evaluator = self.kind.value(data[0], data[1])
        self.__evaluator = evaluator

    def __call__(self, t: float) -> numpy.ndarray:
        return self.__evaluator(t)


class InterpolAssignString:
    """
    Create an executable assignment to handle time boundary conditions of the kind `variable = F(t, data)`,
    where F is an interpolation function of some dataset at time t.
    This class is very similar to `cosapp.core.eval_str.AssignString`, except the right-hand side is a function.
    In order to limit the scope of the callable function, the rhs can only be of type `Interpolator`.
    """
    def __init__(self, variable, function, context):
        # Strict type check (as opposed to `isinstance`), to ensure the type of
        # `function` is `Interpolator`, and is not derived from `Interpolator`.
        if type(function) is not Interpolator:
            raise TypeError(f"Functions used in time boundary conditions may only be of type `Interpolator`; got {type(function)}")
        Boundary.parse(context, variable)  # checks that variable is valid
        fname = f"BC{id(function)}"
        assignment = f"{context.name}.{variable} = float({fname}(t))"
        self.__context = context
        self.__locals = {fname: function, context.name: context, 't': 0}
        self.__code = compile(assignment, "<string>", "exec")  # type: CodeType
        self.__str = f"{variable} = {function.__class__.__name__}(t)"

    def exec(self) -> None:
        """
        Evaluates rhs, and executes assignment lhs <- rhs.
        """
        exec(self.__code, {}, self.locals)

    def __str__(self) -> str:
        return self.__str

    @property
    def locals(self) -> Dict[str, Any]:
        """Dict[str, Any]: Context attributes required to evaluate the string expression."""
        self.__locals['t'] = self.__context.t
        return self.__locals
    
    @property
    def constant(self):
        return False


class Scenario:
    """Class managing boundary and initial conditions for time simulations"""

    def __init__(self, name: str, owner: Driver) -> None:
        """Initialize object

        Parameters
        ----------
        name: str
            Name of the `Module`
        owner : Driver
            :py:class:`~cosapp.drivers.driver.Driver` to which object belong
        """
        self.__case_values = []   # type: List[AssignString]
        self.__init_values = []   # type: List[AssignString]
        self.name = name
        self.owner = owner

    @classmethod
    def make(cls, name: str, init: Dict[str, Any], values: Dict[str, Any], driver: Driver) -> "Scenario":
        """Scenario factory"""
        scenario = cls(name, driver)
        scenario.set_init(init)
        scenario.set_values(values)
        return scenario

    def apply_init_values(self) -> None:
        """Execute assignments corresponding to initial conditions"""
        logger.debug("Apply initial conditions")
        for assignment in self.__init_values:
            logger.debug(f"\t{assignment}")
            assignment.exec()

    def update_values(self) -> None:
        """Execute assignments corresponding to boundary conditions"""
        for assignment in self.__case_values:
            assignment.exec()

    @property
    def context(self) -> "System":
        """System: evaluation context of initial and boundary conditions,
        that is the system controled by owner driver"""
        return self.__context

    @property
    def owner(self) -> Driver:
        """Driver: owner driver"""
        return self.__owner

    @owner.setter
    def owner(self, driver: Driver) -> None:
        check_arg(driver, "owner", Driver, lambda driver: hasattr(driver, "owner"))
        self.__owner = driver
        self.__context = driver.owner
        self.clear_init()
        self.clear_values()

    def add_init(self, modifications: Dict[str, Any]) -> None:
        """Add a set of initial conditions, from a dictionary of the kind {'variable': value, ...}

        Parameters
        ----------
        modifications : Dict[str, Any]
            Dictionary of (variable name, value)

        Examples
        --------
        >>> scenario.add_init({'myvar': 42, 'port.dummy': '-2 * alpha'})
        """
        check_arg(modifications, 'modifications', dict)

        if self.owner is None:
            raise AttributeError(f"Driver {self.name!r} must be attached to a System to set initial values.")

        for variable, value in modifications.items():
            assignment = self.__assignment(variable, value)
            if assignment.constant:
                # If assignment is constant, it is safer to insert it at the top
                # of the list, since other initial condition assignments might use it.
                self.__init_values.insert(0, assignment)
            else:
                self.__init_values.append(assignment)

    def set_init(self, modifications: Dict[str, Any]) -> None:
        """Set initial conditions, from a dictionary of the kind {'variable': value, ...}

        See `add_init` for further detail.
        """
        self.clear_init()
        self.add_init(modifications)

    def add_values(self, modifications: Dict[str, Any]) -> None:
        """Add a set of variables to the list of case values, from a dictionary of the kind {'variable': value, ...}

        Each variable and its value can be contextual, as in {'child1.port2.var': '2 * child2.foo'},
        as long as they are both evaluable in the context of the driver's owner.
        Explicit time dependency can be given using variable 't' in values, as in 'exp(-t / tau)'

        Parameters
        ----------
        modifications : Dict[str, Any]
            Dictionary of (variable name, value)

        Examples
        --------
        >>> scenario.add_values({'myvar': 42, 'port.dummy': 'cos(omega * t)'})
        """
        check_arg(modifications, 'modifications', dict)

        if self.owner is None:
            raise AttributeError(f"Driver {self.name!r} must be attached to a System to set case values.")

        for variable, value in modifications.items():
            if type(value) is Interpolator:
                assignment = InterpolAssignString(variable, value, self.context)
            else:
                assignment = self.__assignment(variable, value)
            if assignment.constant:
                # If assignment is constant, it can be regarded as an initial condition,
                # rather than a time-dependent boundary condition.
                # Moreover, it is safer to insert it at the top of the list,
                # since other initial condition assignments might use it.
                self.__init_values.insert(0, assignment)
            else:
                self.__case_values.append(assignment)

    def set_values(self, modifications: Dict[str, Any]) -> None:
        """Set case values, from a dictionary of the kind {'variable': value, ...}

        See `add_values` for further detail.
        """
        self.clear_values()
        self.add_values(modifications)

    def clear_values(self) -> None:
        """Clears the list of boundary conditions"""
        self.__case_values.clear()

    def clear_init(self) -> None:
        """Clears the list of initial conditions"""
        self.__init_values.clear()

    @property
    def case_values(self) -> List[AssignString]:
        """List[AssignString]: list of boundary conditions"""
        return self.__case_values

    @property
    def init_values(self) -> List[AssignString]:
        """List[AssignString]: list of initial conditions"""
        return self.__init_values

    def __assignment(self, variable, value) -> AssignString:
        """Returns the AssignString corresponding to assignment `variable <- value`"""
        context = self.__context
        Boundary.parse(context, variable)  # checks that variable is valid
        return AssignString(variable, value, context)

    def __repr__(self) -> str:
        s = f"{self.__class__.__qualname__} {self.name!r}, in {self.context.name!r}"
        def conditions(assigments, title):
            output = ""
            prefix = f"\n{title.title()}:\n\t"
            for assignment in assigments:
                output += prefix + str(assignment)
                prefix = "\n\t"
            return output
        s += conditions(self.init_values, "Initial values")
        s += conditions(self.case_values, "Boundary conditions")
        return s



"""
Classes and functions to support unit conversion.

The module provides a basic set of predefined physical quantities
in its built-in library; however, it also supports generation of
personal libraries which can be saved and reused.
This module is based on the PhysicalQuantities module
in Scientific Python, by Konrad Hinsen.

It has been modified by Justin Gray for integration in OpenMDAO 2.2 and by the CoSApp team
for integration in CoSApp (mainly remove Python 2 support).
"""

import os.path
import re
from collections import OrderedDict
from configparser import _UNSET, DEFAULTSECT, RawConfigParser
from math import floor, pi
from numbers import Number
from typing import Any, Dict, List, Optional, Tuple, Union


# TODO could it be more interesting to merge this with the PhysicalQuantities package of
# https://github.com/juhasch/PhysicalQuantities - unfortunately this package as lots of others don't
# have a easy way to get the convertion tuple (offset, factor)...
# TODO this code is not the cleanest nor the robustest
####################################
# Class Definitions
####################################
class UnitError(Exception):
    """Raised if units are incompatible or unknown.

    Parameters
    ----------
    message : str
        Error message

    Attributes
    ----------
    message : str
        Error message
    """

    def __init__(self, message: str):
        """Instantiate a error object from the error descriptive message.

        Parameters
        ----------
        message : str
            Error message
        """
        self.message = message


class NumberDict(OrderedDict):
    """
    Dictionary storing numerical values.

    An instance of this class acts like an array of numbers with
    generalized (non-integer) indices. A value of zero is assumed
    for undefined entries. NumberDict instances support addition
    and subtraction with other NumberDict instances, and multiplication
    and division by scalars.
    """

    def __getitem__(self, item: str) -> int:
        """
        Get the item, or 0.

        Parameters
        ----------
        item : key
            key to get the item

        Returns
        -------
        int
            value of the given key
        """
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            return 0

    def __coerce__(self, other: Dict[str, float]) -> Tuple['NumberDict', 'NumberDict']:
        """
        Change other dict to NumberDicts.

        Parameters
        ----------
        other : Dict
            the dict instance to be coerced

        Returns
        -------
        Tuple[NumberDict, NumberDict]
            old and new NumberDict with keys/values from original
        """
        if isinstance(other, dict):
            other = NumberDict(other)
        return self, other

    def __add__(self, other: 'NumberDict') -> 'NumberDict':
        """
        Add another NumberDict to myself.

        Parameters
        ----------
        other : NumberDict
            the other NumberDict Instance

        Returns
        -------
        NumberDict
            new NumberDict with self+other values
        """
        sum_dict = NumberDict()
        for k, v in self.items():
            sum_dict[k] = v
        for k, v in other.items():
            sum_dict[k] = sum_dict[k] + v
        return sum_dict

    def __sub__(self, other: 'NumberDict') -> 'NumberDict':
        """
        Add another NumberDict from myself.

        Parameters
        ----------
        other : NumberDict
            the other NumberDict Instance

        Returns
        -------
        NumberDict
            new NumberDict instance, with self-other values
        """
        sum_dict = NumberDict()
        for k, v in self.items():
            sum_dict[k] = v
        for k, v in other.items():
            sum_dict[k] = sum_dict[k] - v
        return sum_dict

    def __rsub__(self, other: 'NumberDict') -> 'NumberDict':
        """
        Add subtract myself from another NumberDict.

        Parameters
        ----------
        other : NumberDict
            the other NumberDict Instance

        Returns
        -------
        NumberDict
            new NumberDict instance, with other-self values
        """
        sum_dict = NumberDict()
        for k, v in other.items():
            sum_dict[k] = v
        for k, v in self.items():
            sum_dict[k] = sum_dict[k] - v
        return sum_dict

    def __mul__(self, other: int) -> 'NumberDict':
        """
        Multiply myself by other.

        Parameters
        ----------
        other : int
            value to multiply by

        Returns
        -------
        NumberDict
            new NumberDict instance, with other*self values
        """
        new = NumberDict()
        for key, value in self.items():
            new[key] = other * value
        return new

    __rmul__ = __mul__

    def __div__(self, other: int) -> 'NumberDict':
        """
        Divide myself by other.

        Parameters
        ----------
        other : int
            value to divide by

        Returns
        -------
        NumberDict
            new NumberDict instance, with self/other values
        """
        new = NumberDict()
        for key, value in self.items():
            new[key] = value / other
        return new

    __truediv__ = __div__  # for python 3

    def __repr__(self) -> str:
        """
        Return a string deceleration of myself.

        Returns
        -------
        str
            str representation for the creation of this NumberDict
        """
        return repr(dict(self))


class PhysicalUnit(object):
    """
    Physical unit.

    A physical unit is defined by a name (possibly composite), a scaling
    factor, and the exponentials of each of the SI base units that enter into
    it. Units can be multiplied, divided, and raised to integer powers.

    Attributes
    ----------
    _names : dict or str
        A dictionary mapping each name component to its
        associated integer power (e.g., C{{'m': 1, 's': -1}})
        for M{m/s}). As a shorthand, a string may be passed
        which is assigned an implicit power 1.
    _factor : float
        A scaling factor.
    _powers : list of int
        The integer powers for each of the nine base units.
    _offset : float
        An additive offset to the base unit (used only for temperatures)
    """

    def __init__(self, names: Union[str, Dict[str, int]], factor: float, powers: List[int],
                 offset: float=0):
        """
        Initialize all attributes.

        Parameters
        ----------
        names : dict or str
            A dictionary mapping each name component to its
            associated integer power (e.g., C{{'m': 1, 's': -1}})
            for M{m/s}). As a shorthand, a string may be passed
            which is assigned an implicit power 1.
        factor : float
            A scaling factor.
        powers : list of int
            The integer powers for each of the nine base units.
        offset : float
            An additive offset to the base unit (used only for temperatures).
        """
        if isinstance(names, str):
            self._names = NumberDict(((names, 1),))  # type: NumberDict
        else:
            self._names = NumberDict(names)  # type: NumberDict

        self._factor = float(factor)  # type: float
        self._offset = float(offset)  # type: float
        self._powers = powers  # type: List[int]

    def __repr__(self) -> str:
        """
        Get the string representation of this unit.

        Returns
        -------
        str
            str representation of how to instantiate this PhysicalUnit
        """
        return 'PhysicalUnit(%s,%s,%s,%s)' % (self._names, self._factor,
                                              self._powers, self._offset)

    def __str__(self) -> str:
        """
        Convert myself to string.

        Returns
        -------
        str
            str representation of a PhysicalUnit
        """
        return '<PhysicalUnit ' + self.name() + '>'

    def __lt__(self, other: 'PhysicalUnit') -> bool:
        """
        Compare myself to other.

        Parameters
        ----------
        other : PhysicalUnit
            The other physical unit to be compared to

        Returns
        -------
        bool
            self._factor < other._factor
        """
        if self._powers != other._powers or self._offset != other._offset:
            raise UnitError(f"Unit {other.name()} is not compatible with {self.name()}.")

        return self._factor < other._factor

    def __gt__(self, other: 'PhysicalUnit') -> bool:
        """
        Compare myself to other.

        Parameters
        ----------
        other : PhysicalUnit
            The other physical unit to be compared to

        Returns
        -------
        bool
            self._factor > other._factor
        """
        if self._powers != other._powers:
            raise UnitError(f"Unit {other.name()} is not compatible with {self.name()}.")

        return self._factor > other._factor

    def __eq__(self, other: 'PhysicalUnit') -> bool:
        """
        Test for equality.

        Parameters
        ----------
        other : PhysicalUnit
            The other physical unit to be compared to

        Returns
        -------
        bool
            true if _factor, _offset, and _powers all match
        """
        return (self._factor == other._factor and
                self._offset == other._offset and
                self._powers == other._powers)

    # TODO the case for other is a number seems wrong
    def __mul__(self, other: Union['PhysicalUnit', int]) -> 'PhysicalUnit':
        """
        Multiply myself by other.

        Parameters
        ----------
        other : PhysicalUnit or int
            The other physical unit or the magnitude to be multiplied with

        Returns
        -------
        PhysicalUnit
            new PhysicalUnit instance representing the product of two units
        """
        if abs(self._offset) > 1e-9 or (isinstance(other, PhysicalUnit) and
                                        abs(other._offset) > 1e-9):
            raise UnitError("cannot multiply units with non-zero offset")
        if isinstance(other, PhysicalUnit):
            return PhysicalUnit(self._names + other._names,
                                self._factor * other._factor,
                                [a + b for (a, b) in zip(self._powers,
                                                         other._powers)])
        else:
            return PhysicalUnit(self._names + {str(other): 1},
                                self._factor * other,
                                self._powers,
                                self._offset * other)

    __rmul__ = __mul__

    # TODO the case for other is a number seems wrong
    def __div__(self, other: Union['PhysicalUnit', int]) -> 'PhysicalUnit':
        """
        Divide myself by other.

        Parameters
        ----------
        other : PhysicalUnit or int
            The other physical unit or magnitude to be divided by

        Returns
        -------
        PhysicalUnit
            new PhysicalUnit instance representing the self/other
        """
        if abs(self._offset) > 1e-9 or (isinstance(other, PhysicalUnit) and
                                        abs(other._offset) > 1e-9):
            raise UnitError("cannot divide units with non-zero offset")
        if isinstance(other, PhysicalUnit):
            return PhysicalUnit(self._names - other._names,
                                self._factor / other._factor,
                                [a - b for (a, b) in zip(self._powers,
                                                         other._powers)])
        else:
            return PhysicalUnit(self._names + {str(other): -1},
                                self._factor / float(other), self._powers)

    __truediv__ = __div__   # for python 3

    # TODO this seems wrong
    def __rdiv__(self, other: int) -> 'PhysicalUnit':
        """
        Divide other by myself.

        Parameters
        ----------
        other : int
            The other value to be divided by this unit

        Returns
        -------
        PhysicalUnit
            new PhysicalUnit instance representing the other/self
        """
        return PhysicalUnit({str(other): 1} - self._names,
                            float(other) / self._factor,
                            [-x for x in self._powers])

    __rtruediv__ = __rdiv__

    def __pow__(self, other: Union[int, float]) -> 'PhysicalUnit':
        """
        Raise myself to a power.

        Parameters
        ----------
        other : float or int
            power to raise self by

        Returns
        -------
        PhysicalUnit
            new PhysicalUnit of self^other
        """
        if abs(self._offset) > 1e-9:
            raise UnitError("cannot exponentiate units with non-zero offset")
        if isinstance(other, int):  # TODO something is wrong here
            return PhysicalUnit(self._names * other, pow(self._factor, other),
                                [x * other for x in self._powers])
        if isinstance(other, float):
            inv_exp = 1. / other
            rounded = int(floor(inv_exp + 0.5))
            if abs(inv_exp - rounded) < 1.e-10:

                if all([x % rounded == 0 for x in self._powers]):
                    f = self._factor**other
                    p = [x / rounded for x in self._powers]
                    if all([x % rounded == 0 for x in self._names.values()]):
                        names = self._names / rounded
                    else:
                        names = NumberDict()
                        if f != 1.:
                            names[str(f)] = 1
                        for x, name in zip(p, _UNIT_LIB.base_names):
                            names[name] = x
                    return PhysicalUnit(names, f, p)

        raise TypeError('Only integer and inverse integer exponents allowed')

    def in_base_units(self) -> 'PhysicalUnit':
        """
        Return the base unit equivalent of this unit.

        Returns
        -------
        PhysicalUnit
            the equivalent base unit
        """
        num = ''
        denom = ''
        for unit, power in zip(_UNIT_LIB.base_names, self._powers):
            if power < 0:
                denom = denom + '/' + unit
                if power < -1:
                    denom = denom + '**' + str(-power)
            elif power > 0:
                num = num + '*' + unit
                if power > 1:
                    num = num + '**' + str(power)

        if len(num) == 0:
            num = '1'
        else:
            num = num[1:]

        return _find_unit(num + denom)

    def conversion_tuple_to(self, other: 'PhysicalUnit') -> Tuple[float, float]:
        """
        Compute the tuple of (factor, offset) for conversion.

        Parameters
        ----------
        other : PhysicalUnit
            Another unit.

        Returns
        -------
        Tuple with two floats
            The conversion factor and offset from this unit to another unit.
        """
        if self._powers != other._powers:
            raise UnitError(f"Unit {other.name()} is not compatible with {self.name()}.")

        # let (s1,d1) be the conversion tuple from 'self' to base units
        #   (ie. (x+d1)*s1 converts a value x from 'self' to base units,
        #   and (x/s1)-d1 converts x from base to 'self' units)
        # and (s2,d2) be the conversion tuple from 'other' to base units
        # then we want to compute the conversion tuple (S,D) from
        #   'self' to 'other' such that (x+D)*S converts x from 'self'
        #   units to 'other' units
        # the formula to convert x from 'self' to 'other' units via the
        #   base units is (by definition of the conversion tuples):
        #     ( ((x+d1)*s1) / s2 ) - d2
        #   = ( (x+d1) * s1/s2) - d2
        #   = ( (x+d1) * s1/s2 ) - (d2*s2/s1) * s1/s2
        #   = ( (x+d1) - (d1*s2/s1) ) * s1/s2
        #   = (x + d1 - d2*s2/s1) * s1/s2
        # thus, D = d1 - d2*s2/s1 and S = s1/s2

        factor = self._factor / other._factor
        offset = self._offset - (other._offset * other._factor / self._factor)
        return factor, offset

    def is_compatible(self, other: 'PhysicalUnit') -> bool:
        """
        Check for compatibility with another unit.

        Parameters
        ----------
        other : PhysicalUnit
            Another unit.

        Returns
        -------
        bool
            indicates if two units are compatible
        """
        return self._powers == other._powers

    def is_dimensionless(self) -> bool:
        """
        Dimensionless PQ.

        Returns
        -------
        bool
            indicates if this is dimensionless
        """
        return not any(self._powers)

    def is_angle(self) -> bool:
        """
        Check if this PQ is an Angle.

        Returns
        -------
        bool
            indicates if this an angle type
        """
        return (self._powers[_UNIT_LIB.base_types['angle']] == 1 and
                sum(self._powers) == 1)

    def set_name(self, name: str):
        """
        Set the name.

        Parameters
        ----------
        name : str
            the name
        """
        self._names = NumberDict()
        self._names[name] = 1

    def name(self) -> str:
        """
        Compute the name of this unit.

        Returns
        -------
        str
            str representation of the unit
        """
        num = ''
        denom = ''
        for unit, power in self._names.items():
            if power < 0:
                denom = denom + '/' + unit
                if power < -1:
                    denom = denom + '**' + str(-power)
            elif power > 0:
                num = num + '*' + unit
                if power > 1:
                    num = num + '**' + str(power)
        if len(num) == 0:
            num = '1'
        else:
            num = num[1:]
        return num + denom


####################################
# Module Functions
####################################

def _new_unit(name: str, factor: float, powers: List[int]) -> None:
    """
    Create new PhysicalUnit and add it to the unit library.

    Parameters
    ----------
    name : str
        The name of the new unit
    factor : float
        conversion factor to base units
    powers : list of int
        power of base units
    """
    _UNIT_LIB.unit_table[name] = PhysicalUnit(name, factor, powers)


def add_offset_unit(name: str, baseunit: Union[str, PhysicalUnit], factor: float, offset: float,
                    comment: str='') -> None:
    """
    Adding Offset PhysicalUnit to the unit library.

    Parameters
    ----------
    name : str
        The name of the unit
    baseunit : str or PhysicalUnit
        The unit upon which this offset unit is based.
    factor : float
        The scaling factor used to define the new unit w.r.t. baseunit
    offset : float
        zero offset for new unit
    comment : str
        optional comment to describe unit
    """
    if isinstance(baseunit, str):
        baseunit = _find_unit(baseunit)
    # else, baseunit should be a instance of PhysicalUnit
    # names, factor, powers, offset=0
    unit = PhysicalUnit(baseunit._names, baseunit._factor * factor,
                        baseunit._powers, offset)
    unit.set_name(name)
    if name in _UNIT_LIB.unit_table:
        if (_UNIT_LIB.unit_table[name]._factor != unit._factor or
                _UNIT_LIB.unit_table[name]._powers != unit._powers):
            raise KeyError("Unit %s already defined with " % name +
                           "different factor or powers")
    _UNIT_LIB.unit_table[name] = unit
    _UNIT_LIB.set('units', name, unit)
    if comment:
        _UNIT_LIB.help.append((name, comment, unit))


def add_unit(name: str, unit: str, comment: str = '') -> None:
    """
    Adding PhysicalUnit from its name and string representation to the unit library.

    Parameters
    ----------
    name : str
        The name of the unit being added. For example: 'Hz'
    unit : str
        definition of the unit w.r.t. some other unit.  For example: '1/s'
    comment : str
        optional comment to describe unit
    """
    if comment:
        _UNIT_LIB.help.append((name, comment, unit))
    if isinstance(unit, str):
        unit = eval(unit, {'__builtins__': None, 'pi': pi},
                    _UNIT_LIB.unit_table)
    unit.set_name(name)
    if name in _UNIT_LIB.unit_table:
        if (_UNIT_LIB.unit_table[name]._factor != unit._factor or
                _UNIT_LIB.unit_table[name]._powers != unit._powers):
            raise KeyError("Unit %s already defined with " % name +
                           "different factor or powers")

    _UNIT_LIB.unit_table[name] = unit
    _UNIT_LIB.set('units', name, unit)


class UnitConfigParser(RawConfigParser):
    """Customized configuration file parser to store physical units."""

    def __init__(self, defaults=None, dict_type=OrderedDict, allow_no_value=False,
                 delimiters=('=', ':'), comment_prefixes=('#', ';'), inline_comment_prefixes=None,
                 strict=True, empty_lines_in_values=True, default_section=DEFAULTSECT,
                 interpolation=_UNSET):
        super().__init__(defaults, dict_type, allow_no_value, delimiters=delimiters,
                         comment_prefixes=comment_prefixes,
                         inline_comment_prefixes=inline_comment_prefixes, strict=strict,
                         empty_lines_in_values=empty_lines_in_values,
                         default_section=default_section, interpolation=interpolation)

        self.base_names = list()  # type: List[str]
        # used to is_angle() and other base type checking
        self.base_types = dict()  # type: Dict[str, int]
        self.unit_table = dict()
        self.prefixes = dict()  # type: Dict[str, float]
        self.help = list()

    def optionxform(self, optionstr: str) -> str:
        """
        Make the RawConfigParser case sensitive.

        Defines an optionxform for the units configparser that
        does nothing, resulting in a case-sensitive parser.

        Parameters
        ----------
        optionstr : str
            The string to be transformed for the RawConfigParser

        Returns
        -------
        str
            The same string that was given as a parameter.
        """
        return optionstr


_UNIT_LIB = UnitConfigParser()


def import_library(libfilepointer) -> UnitConfigParser:
    """
    Import a units library, replacing any existing definitions.

    Parameters
    ----------
    libfilepointer : file
        new library file to work with

    Returns
    -------
    UnitConfigParser
        newly updated units library for the module
    """
    global _UNIT_LIB
    global _UNIT_CACHE
    _UNIT_CACHE = {}
    _UNIT_LIB = UnitConfigParser()
    _UNIT_LIB.read_file(libfilepointer)
    required_base_types = ['length', 'mass', 'time', 'temperature', 'angle']

    for prefix, factor in _UNIT_LIB.items('prefixes'):
        factor, comma, comment = factor.partition(',')
        _UNIT_LIB.prefixes[prefix] = float(factor)

    base_list = [0] * len(_UNIT_LIB.items('base_units'))

    for i, (unit_type, name) in enumerate(_UNIT_LIB.items('base_units')):
        _UNIT_LIB.base_types[unit_type] = i
        powers = list(base_list)
        powers[i] = 1
        # print '%20s'%unit_type, powers
        # cant use add_unit because no base units exist yet
        _new_unit(name, 1, powers)
        _UNIT_LIB.base_names.append(name)

    # test for required base types
    missing = [utype for utype in required_base_types
               if utype not in _UNIT_LIB.base_types]
    if missing:
        raise ValueError('Not all required base type were present in the'
                         ' config file. missing: %s, at least %s required'
                         % (missing, required_base_types))

    # Explicit unitless 'unit'.
    _new_unit('unitless', 1, list(base_list))
    _update_library(_UNIT_LIB)
    return _UNIT_LIB


def update_library(filename: Union[str, Any]) -> None:
    """
    Update units in current library from `filename`.

    Parameters
    ----------
    filename : string or file
        Source of units configuration data.
    """
    if isinstance(filename, str):
        inp = open(filename, 'rU')
    else:
        inp = filename
    try:
        cfg = RawConfigParser()
        cfg.optionxform = lambda s: s
        cfg.read_file(inp)
        _update_library(cfg)
    finally:
        inp.close()


def _update_library(cfg: RawConfigParser) -> None:
    """
    Update library from :class:`RawConfigParser` `cfg`.

    Parameters
    ----------
    cfg : RawConfigParser
        RawConfigParser loaded with unit_lib.ini data
    """
    retry1 = set()
    for name, unit in cfg.items('units'):
        data = [item.strip() for item in unit.split(',')]
        if len(data) == 2:
            unit, comment = data
            try:
                add_unit(name, unit, comment)
            except NameError:
                retry1.add((name, unit, comment))
        elif len(data) == 4:
            factor, baseunit, offset, comment = data
            try:
                add_offset_unit(name, baseunit, float(factor), float(offset),
                                comment)
            except NameError:
                retry1.add((name, baseunit, float(factor), float(offset),
                            comment))
        else:
            raise ValueError(f"Unit {name!r} definition {unit!r} has invalid format")
    retry_count = 0
    last_retry_count = -1
    while last_retry_count != retry_count and retry1:
        last_retry_count = retry_count
        retry_count = 0
        retry2 = retry1.copy()
        for data in retry2:
            if len(data) == 3:
                name, unit, comment = data
                try:
                    add_unit(name, unit, comment)
                    retry1.remove(data)
                except NameError:
                    retry_count += 1
            else:
                try:
                    name, factor, baseunit, offset, comment = data
                    add_offset_unit(name, factor, baseunit, offset, comment)
                    retry1.remove(data)
                except NameError:
                    retry_count += 1
    if retry1:
        raise ValueError('The following units were not defined because they'
                         ' could not be resolved as a function of any other'
                         ' defined units:%s' % [x[0] for x in retry1])


_UNIT_CACHE = {}  # type: Dict[str, PhysicalUnit]


def _find_unit(unit: Union[str, PhysicalUnit, None]) -> Optional[PhysicalUnit]:
    """
    Find unit helper function.

    Parameters
    ----------
    unit : str or PhysicalUnit
        str representing the desired unit

    Returns
    -------
    PhysicalUnit or None
        The actual unit object - None if not found
    """
    if isinstance(unit, str):
        name = unit.strip()
        try:
            unit = _UNIT_CACHE[name]
        except KeyError:
            try:
                unit = eval(name, {'__builtins__': None}, _UNIT_LIB.unit_table)
            except (SyntaxError, NameError, TypeError):

                # This unit might include prefixed units that aren't in the
                # unit_table. We must parse them ALL and add them to the
                # unit_table.

                # First character of a unit is always alphabet.
                # Remaining characters may include numbers.
                regex = re.compile('[A-z][A-z0-9]*')

                if not regex.match(name):
                    return None

                for item in regex.findall(name):
                    # check if this was a compound unit, so each
                    # substring might be a unit
                    try:
                        eval(item, {'__builtins__': None}, _UNIT_LIB.unit_table)
                    except (SyntaxError, NameError, TypeError):  # maybe is a prefixed unit then
                        # check for single letter prefix before unit
                        if(item[0] in _UNIT_LIB.prefixes and
                           item[1:] in _UNIT_LIB.unit_table):
                            add_unit(item, _UNIT_LIB.prefixes[item[0]] *
                                     _UNIT_LIB.unit_table[item[1:]])

                        # check for double letter prefix before unit
                        elif(item[0:2] in _UNIT_LIB.prefixes and
                             item[2:] in _UNIT_LIB.unit_table):
                            add_unit(item, _UNIT_LIB.prefixes[item[0:2]] *
                                     _UNIT_LIB.unit_table[item[2:]])

                        # no prefixes found, unknown unit
                        else:
                            return None
                            # raise ValueError("no unit named '%s' is defined"
                            #                  % item)

                unit = eval(name, {'__builtins__': None}, _UNIT_LIB.unit_table)

            _UNIT_CACHE[name] = unit

    if not isinstance(unit, PhysicalUnit):
        return None

    return unit


def is_valid_units(unit: Union[str, PhysicalUnit]) -> bool:
    """
    Return whether the given units are valid.

    Parameters
    ----------
    unit : str or PhysicalUnit
        Unit to test for.

    Returns
    -------
    bool
        True for valid, False for invalid.
    """
    return _find_unit(unit) is not None


def conversion_to_base_units(units: Union[str, PhysicalUnit]) -> Tuple[float, float]:
    """
    Get the offset and scaler to convert from given units to base units.

    Parameters
    ----------
    units : str or PhysicalUnit
        Unit or its representation.

    Returns
    -------
    float
        Offset to get to default unit: m (length), s(time), etc.
    float
        Mult. factor to get to default unit: m (length), s(time), etc.
    """
    if not units:  # dimensionless
        return 0., 1.
    unit = _find_unit(units)

    return unit._offset, unit._factor


def is_compatible(old_units: Union[str, PhysicalUnit], new_units: Union[str, PhysicalUnit]) -> bool:
    """
    Check whether units are compatible in terms of base units.

    e.g., m/s is compatible with ft/hr

    Parameters
    ----------
    old_units : str or PhysicalUnit
        original units
    new_units : str or PhysicalUnit
        new units to check compatibility with

    Returns
    -------
    bool
        whether the units are compatible.
    """
    if not old_units and not new_units:  # dimensionless
        return True

    old_unit = _find_unit(old_units)
    new_unit = _find_unit(new_units)

    if old_unit is None or new_unit is None:
        return False

    return old_unit.is_compatible(new_unit)


def get_conversion(old_units: Union[str, PhysicalUnit, None], new_units: Union[str, PhysicalUnit, None]) \
        -> Optional[Tuple[float, float]]:
    """
    Return conversion factor and offset between old and new units.

    Parameters
    ----------
    old_units : str or PhysicalUnit
        original units
    new_units : str or PhysicalUnit
        new units to get the conversion from.

    Returns
    -------
    (float, float) or None
        Conversion factor and offset or None is no conversion available
    """
    old_unit = _find_unit(old_units)
    new_unit = _find_unit(new_units)

    if old_unit is None or new_unit is None:
        return None
    else:
        return old_unit.conversion_tuple_to(new_unit)


def convert_units(val: float,
                  old_units: Union[str, PhysicalUnit, None] = None,
                  new_units: Union[str, PhysicalUnit, None] = None) -> float:
    """
    Take a given quantity and return in different units.

    If there is no current units or no target units, the original value is returned.

    Parameters
    ----------
    val : Number
        value in original units.
    old_units : str or PhysicalUnit or None
        original units or None.
    new_units : str or PhysicalUnit or None
        new units to return the value in or None.

    Returns
    -------
    float
        value in new units.
    """
    if not old_units or not new_units:  # one side has no units
        return val

    old_unit = _find_unit(old_units)
    new_unit = _find_unit(new_units)

    (factor, offset) = old_unit.conversion_tuple_to(new_unit)
    return (val + offset) * factor


# Load in the default unit library
with open(os.path.join(os.path.dirname(__file__), 'unit_library.ini')) as default_lib:
    import_library(default_lib)


if __name__ == '__main__':
    for returned, expected in [
        (conversion_to_base_units('cm'), (0., 1.0e-2)),
        (conversion_to_base_units('km'), (0., 1.0e3)),
        (convert_units(3.0, 'mm'), 3.0e-3),
        (convert_units(3.0, 'mm', 'cm'), 3.0e-1),
        (convert_units(100, 'degC', 'degF'), 212.)
    ]:
        print(returned, 'should be', expected)



import pytest

from cosapp.core.eval_str import ContextLocals
from cosapp.systems.system import System


def test_ContextLocals___init___():
    a = System('dummy')
    local = ContextLocals(a)
    assert len(local) == 0
    assert local.context is a


def test_ContextLocals_missing_keys_success():
    class Dummy(System):
        def setup(self):
            self.add_inward('x', 1)
        
        def hello(self):
            print('hello')

    a = Dummy('dummy')
    local = ContextLocals(a)
    assert len(local) == 0
    assert local['x'] == a.x
    assert len(local) == 1
    assert local['hello'] == a.hello
    assert len(local) == 2
    assert set(local.keys()) == set(('x', 'hello'))


def test_ContextLocals_missing_keys_failure():
    a = System('dummy')
    local = ContextLocals(a)

    with pytest.raises(KeyError):
        local['a']



"""Class defining an uniform distribution."""
from typing import Optional

import numpy
import scipy.stats

from .distribution import Distribution


class Uniform(Distribution):
    """A class to define an uniform distribution.

    Parameters
    ----------
    worst : float
        The parameter absolute worst value.
    best : float
        The parameter absolute best value.
    pworst : float, optional
        The worst value probability is the probability that the variable will be lower (if worst<best) or higher
        (if worst>best) than the worst value; default 0.15 (i.e. 15%).
    pbest : float, optional
        The best value probability is the probability that the variable will be higher (if worst<best) or lower
        (if worst>best) than the best value; default 0.15 (i.e. 15%).
   """

    def __init__(
        self,
        worst: float,
        best: float,
        pworst: Optional[float] = 0.15,
        pbest: Optional[float] = 0.15,
    ):
        self._rv = None  # type: scipy.stats.uniform

        super().__init__(worst, best, pworst, pbest)

    def _set_distribution(self):
        """Set the probability distribution according the parameters."""
        if self.pworst + self.pbest > 1:
            raise ValueError(
                f"Best and worst probabilities are incompatible: {self.__json__()!s}."
            )

        scale = numpy.abs(self.worst - self.best) / (1 - self.pworst - self.pbest)
        if self.worst < self.best:
            loc = self.worst - scale * self.pworst
        else:
            loc = self.best - scale * self.pbest

        self._rv = scipy.stats.uniform(loc=loc, scale=scale)

    def draw(self, quantile: Optional[float] = None) -> float:
        """Generate a random number.

        If a quantile is given, generate the perturbation for that quantile.

        Parameters
        ----------
        quantile : Optional[float], optional
            Quantile for which the perturbation must be set; default None (i.e. random perturbation)
        
        Returns
        -------
        float
            The random number
        """
        return self._rv.rvs() if quantile is None else self._rv.ppf(quantile)
